# Authors: Luis Scoccola
# License: 3-clause BSD

import unittest
from persistable import Persistable
from persistable.persistable import _HierarchicalClustering, _MetricProbabilitySpace
from scipy.spatial import distance_matrix
from scipy.spatial.distance import cdist
from sklearn import datasets
import numpy as np


class TestMetricProbabilitySpace(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        self._n = 25
        self._X = np.random.random_sample((self._n, 2))
        self._ps = [2, 4, 8, np.inf]
        self._s0s = np.linspace(0.1, 2, 10)
        self._k0s = np.linspace(0.01, 0.5, 10)
        self._number_different_weights = 3
        self._different_weights = [None]
        for _ in range(self._number_different_weights):
            self._different_weights.append(np.random.random_sample(self._n))

    def test_core_distances(self):
        n = 4
        X = np.array([[0, 0], [1, 0], [1, 1], [3, 0]])
        p = Persistable(X)
        mps = p._mpspace

        s0 = 2
        k0 = 0.5
        res = np.array([1, 1, 1, 1])
        np.testing.assert_almost_equal(mps._core_distance(np.arange(n), s0, k0), res)

        s0 = 1
        k0 = 0.5
        res = np.array([1 / 2, 1 / 2, 1 / 2, 1 / 2])
        np.testing.assert_almost_equal(mps._core_distance(np.arange(n), s0, k0), res)

        s0 = np.infty
        k0 = 1
        res = np.array([3, 2, np.sqrt(5), 3])
        np.testing.assert_almost_equal(mps._core_distance(np.arange(n), s0, k0), res)

        s0 = np.infty
        k0 = 0.5
        res = np.array([1, 1, 1, 2])
        np.testing.assert_almost_equal(mps._core_distance(np.arange(n), s0, k0), res)

        p = Persistable(X, measure=np.array([0.5, 0.5, 0.5, 0.5]))
        mps = p._mpspace
        s0 = np.infty
        k0 = 0.6
        res = np.array([1, 1, 1, 2])
        np.testing.assert_almost_equal(mps._core_distance(np.arange(n), s0, k0), res)

    def test_same_core_distances(self):
        for w in self._different_weights:
            for p in self._ps:
                p1 = Persistable(self._X, measure=w, p=p)
                mps1 = p1._mpspace
                p2 = Persistable(
                    distance_matrix(self._X, self._X, p=p),
                    metric="precomputed",
                    measure=w,
                )
                mps2 = p2._mpspace
                for s0 in self._s0s:
                    for k0 in self._k0s:
                        np.testing.assert_almost_equal(
                            mps1._core_distance(np.arange(self._n), s0, k0),
                            mps2._core_distance(np.arange(self._n), s0, k0),
                        )

    def test_same_hierarchy(self):
        # test Boruvka, Prim, and dense Prim
        for w in self._different_weights:
            V = np.ones(self._X.shape[1])
            # will use BallTree and Boruvka
            p1 = Persistable(self._X, metric="seuclidean", measure=w, V=V)
            mps1 = p1._mpspace
            # will use dense MST
            p2 = Persistable(
                cdist(self._X, self._X, metric="seuclidean", V=V),
                metric="precomputed",
                measure=w,
            )
            mps2 = p2._mpspace
            num_components = 1000
            big_X = np.zeros((self._X.shape[0], num_components))
            big_X[:, : self._X.shape[1]] = self._X
            # will use BallTree and Prim
            V2 = np.ones(big_X.shape[1])
            p3 = Persistable(big_X, metric="seuclidean", measure=w, V=V2)
            mps3 = p3._mpspace
            for s0 in self._s0s:
                for k0 in self._k0s:
                    hc1 = mps1.lambda_linkage([0, k0], [s0, 0])
                    hc2 = mps2.lambda_linkage([0, k0], [s0, 0])
                    hc3 = mps3.lambda_linkage([0, k0], [s0, 0])
                    np.testing.assert_almost_equal(
                        hc1._merges_heights, hc2._merges_heights
                    )
                    np.testing.assert_almost_equal(
                        hc2._merges_heights, hc3._merges_heights
                    )
            for p in self._ps:
                # will use KDTree and Boruvka
                p1 = Persistable(self._X, measure=w, p=p)
                mps1 = p1._mpspace
                # will use dense MST
                p2 = Persistable(
                    distance_matrix(self._X, self._X, p=p),
                    metric="precomputed",
                    measure=w,
                )
                mps2 = p2._mpspace
                num_components = 1000
                big_X = np.zeros((self._X.shape[0], num_components))
                big_X[:, : self._X.shape[1]] = self._X
                # will use KDTree and Prim
                p3 = Persistable(big_X, measure=w, p=p)
                mps3 = p3._mpspace
                for s0 in self._s0s:
                    for k0 in self._k0s:
                        hc1 = mps1.lambda_linkage([0, k0], [s0, 0])
                        hc2 = mps2.lambda_linkage([0, k0], [s0, 0])
                        hc3 = mps3.lambda_linkage([0, k0], [s0, 0])
                        np.testing.assert_almost_equal(
                            hc1._merges_heights, hc2._merges_heights
                        )
                        np.testing.assert_almost_equal(
                            hc2._merges_heights, hc3._merges_heights
                        )

    def test_hilbert_function(self):
        X = np.array([[0, 0], [1, 0], [1, 1], [3, 0]])
        p = Persistable(X)
        mps = p._mpspace

        ss = [0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4]
        ks = [0, 1 / 4 + 0.01, 1 / 2, 3 / 4, 1, 1.1]

        res = np.array(
            [
                [4, 4, 2, 2, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 0, 0, 1, 1, 1, 1],
            ]
        )
        np.testing.assert_almost_equal(mps.hilbert_function(ks, ss, n_jobs=4), res)


class TestHierarchicalClustering(unittest.TestCase):
    def clustering_matrix(self, c):
        mat = np.full((c.shape[0], c.shape[0]), -1)
        for i in range(c.shape[0]):
            for j in range(c.shape[0]):
                if c[i] == c[j] and c[i] != -1:
                    mat[i, j] = 0
        return mat

    def test_persistence_diagram(self):
        heights = np.array([0, 1, 3, 2])
        merges = np.array([[0, 1], [2, 4]])
        merges_heights = np.array([2, 6])
        end = 10
        hc = _HierarchicalClustering(heights, merges, merges_heights, 0, end)
        pd = hc.persistence_diagram()
        res = np.array([[0, 10], [1, 2], [3, 6], [2, 10]])
        np.testing.assert_array_equal(
            pd[np.lexsort(pd.T[::-1])], res[np.lexsort(res.T[::-1])]
        )

        heights = np.array([0, 1, 3, 2, 4])
        merges = np.array([[0, 1], [3, 5], [2, 6]])
        merges_heights = np.array([2, 4, 6])
        end = 10
        hc = _HierarchicalClustering(heights, merges, merges_heights, 0, end)
        pd = hc.persistence_diagram()
        res = np.array([[0, 10], [1, 2], [2, 4], [3, 6], [4, 10]])
        np.testing.assert_array_equal(
            pd[np.lexsort(pd.T[::-1])], res[np.lexsort(res.T[::-1])]
        )

        heights = np.array([0.5, 1, 2, 0])
        merges = np.array([[0, 1], [2, 4], [3, 5]])
        merges_heights = np.array([2, 4, 6])
        end = 10
        hc = _HierarchicalClustering(heights, merges, merges_heights, 0, end)
        pd = hc.persistence_diagram()
        res = np.array([[0.5, 6], [1, 2], [2, 4], [0, 10]])
        np.testing.assert_array_equal(
            pd[np.lexsort(pd.T[::-1])], res[np.lexsort(res.T[::-1])]
        )

        heights = np.array([0, 1])
        merges = np.array([[0, 1]])
        merges_heights = np.array([1])
        end = 10
        hc = _HierarchicalClustering(heights, merges, merges_heights, 0, end)
        pd = hc.persistence_diagram()
        res = np.array([[0, 10]])
        np.testing.assert_array_equal(
            pd[np.lexsort(pd.T[::-1])], res[np.lexsort(res.T[::-1])]
        )

    def test_flattening(self):
        heights = np.array([0, 1, 3, 8])
        merges = np.array([[0, 1], [2, 4]])
        merges_heights = np.array([2, 6])
        end = 10
        hc = _HierarchicalClustering(heights, merges, merges_heights, 0, end)
        c = hc.persistence_based_flattening(0.5)
        res = np.array([0, 1, 2, 3])
        np.testing.assert_array_equal(
            self.clustering_matrix(c), self.clustering_matrix(res)
        )

        c = hc.persistence_based_flattening(1.5)
        res = np.array([0, 0, 1, 2])
        np.testing.assert_array_equal(
            self.clustering_matrix(c), self.clustering_matrix(res)
        )


class TestPersistable(unittest.TestCase):
    def test_number_clusters(self):
        n_datapoints = 1000
        n_true_points = int(n_datapoints * 0.7)
        X, _ = datasets.make_blobs(
            n_samples=n_true_points,
            centers=6,
            cluster_std=[0.05, 0.06, 0.07, 0.08, 0.09, 0.1],
            random_state=0,
        )
        np.random.seed(0)
        n_noise = n_datapoints - n_true_points
        noise = (np.random.random_sample((n_noise, 2)) - 0.4) * 4
        X = np.vstack((X, noise))

        p = Persistable(X)
        k0 = 0.05
        for s0 in np.linspace(0.1, 0.5, 5):
            for i in list(range(2, 5)):
                c = p.cluster(n_clusters=i, start=[0, k0], end=[s0, 0])
                self.assertEqual(len(set(c[c >= 0])), i)
                c = p.cluster(
                    n_clusters=i,
                    start=[0, k0],
                    end=[s0, 0],
                    extend_clustering_by_hill_climbing=True,
                )
                self.assertEqual(len(set(c[c >= 0])), i)

    def test_number_clusters_quick_cluster(self):
        X, _ = datasets.make_blobs(
            n_samples=1000, centers=3, cluster_std=[0.05, 0.06, 0.07], random_state=1
        )
        p = Persistable(X)
        c = p.quick_cluster()
        self.assertEqual(len(set(c[c >= 0])), 3)
        c = p.quick_cluster(extend_clustering_by_hill_climbing=True)
        self.assertEqual(len(set(c[c >= 0])), 3)

        X, _ = datasets.make_blobs(n_samples=1000, centers=4, random_state=2)
        p = Persistable(X)
        c = p.quick_cluster(n_neighbors=50)
        self.assertEqual(len(set(c[c >= 0])), 4)
        c = p.quick_cluster(n_neighbors=50, extend_clustering_by_hill_climbing=True)
        self.assertEqual(len(set(c[c >= 0])), 4)

        X, _ = datasets.make_blobs(n_samples=1000, centers=5, random_state=3)
        p = Persistable(X)
        c = p.quick_cluster()
        self.assertEqual(len(set(c[c >= 0])), 5)
        c = p.quick_cluster(extend_clustering_by_hill_climbing=True)
        self.assertEqual(len(set(c[c >= 0])), 5)

    def test_hilbert_function(self):
        X = np.array([[0, 0], [1, 0], [1, 1], [3, 0]])
        p = Persistable(X)

        res_ss = [0.0, 0.5625, 1.125, 1.6875, 2.25, 2.8125, 3.375, 3.9375, 4.5]
        res_ks = [
            0.0,
            0.140625,
            0.28125,
            0.421875,
            0.5625,
            0.703125,
            0.84375,
            0.984375,
            1.125,
        ]

        res = np.array(
            [
                [4, 4, 2, 2, 1, 1, 1, 1],
                [4, 4, 2, 2, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 1, 1, 1, 1, 1, 1],
                [0, 0, 0, 0, 1, 1, 1, 1],
                [0, 0, 0, 0, 1, 1, 1, 1],
            ]
        )
        ss, ks, hs = p._compute_hilbert_function(0, 1, 0, 4, granularity=8)

        np.testing.assert_almost_equal(ss, np.array(res_ss))
        np.testing.assert_almost_equal(ks, np.array(res_ks))
        np.testing.assert_almost_equal(hs, res)

    def test_prominence_vineyard(self):
        X = np.array([[0, 0], [1, 0], [1, 1], [3, 0]])
        p = Persistable(X)

        start_end1 = [(0, 0.1), (10, 0)]
        start_end2 = [(0, 1), (10, 0.9)]
        vineyard = p._compute_vineyard(start_end1, start_end2, n_parameters=4)

        vines = vineyard._vineyard_to_vines()
        res_vines = [
            (np.array([0, 1, 2, 3]), np.array([10.0, 9.0, 9.0, 8.0])),
            (np.array([0, 1, 2, 3]), np.array([2.0, 0.0, 0.0, 0.0])),
            (np.array([0, 1, 2, 3]), np.array([1.0, 0.0, 0.0, 0.0])),
            (np.array([0, 1, 2, 3]), np.array([1.0, 0.0, 0.0, 0.0])),
        ]

        res_vine_parts = [
            [(np.array([0, 10.0, 9.0, 9.0, 8.0]), np.array([3, 0, 1, 2, 3]))],
            [(np.array([0, 2.0, 0.0]), np.array([3, 0, 1]))],
            [(np.array([0, 1.0, 0.0]), np.array([3, 0, 1]))],
            [(np.array([0, 1.0, 0.0]), np.array([3, 0, 1]))],
        ]

        for tv, res_tv, res_vp in zip(vines, res_vines, res_vine_parts):
            t, v = tv
            res_t, res_v = res_tv
            np.testing.assert_almost_equal(t, res_t)
            np.testing.assert_almost_equal(v, res_v)
            np.testing.assert_almost_equal(vineyard._vine_parts(v), res_vp)


if __name__ == "__main__":
    unittest.main()



# cython: boundscheck=False, wraparound=False, cdivision=True

# Copyright (c) 2001-2002 Enthought, Inc. 2003-2022, SciPy Developers.
#
# Function to compute stepwise dendrogram of a hierarchy given by a distance
# matrix plus core distances. modified to be used in Persistable pipeline.
# Originally in scipy/scipy/cluster/_hierarchy.pyx in
# https://github.com/scipy/scipy/commit/2f4bd4234370adcd7085bcc145f9c302f231408f

import numpy as np
cimport numpy as np

from .relabel_dendrogram import label


cdef extern from "numpy/npy_math.h":
    cdef enum:
        NPY_INFINITYF

def stepwise_dendrogram_with_core_distances(int n, const double[:,:] dists, const double[:] core_distance):
    Z_arr = np.empty((n - 1, 3))
    cdef double[:, :] Z = Z_arr

    # Which nodes were already merged.
    cdef int[:] merged = np.zeros(n, dtype=np.intc)

    cdef double[:] D = np.empty(n)
    D[:] = NPY_INFINITYF

    cdef int i, k, x, y = 0
    cdef double dist, current_min

    cdef np.ndarray result_arr 
    cdef np.ndarray core_distance_arr
    cdef np.double_t *core_distance_ptr

    core_distance_arr = np.asarray(core_distance,dtype=np.double)
    cdef np.double_t[::1] core_distance_ = (<np.double_t[:n:1]> (
        <np.double_t *> core_distance_arr.data))

    core_distance_ptr = <np.double_t *> &core_distance_[0]

    x = 0
    with nogil:
        for k in range(n - 1):
            current_min = NPY_INFINITYF
            merged[x] = 1
            for i in range(n):
                if merged[i] == 1:
                    continue

                #dist = max(dists[condensed_index(n, x, i)],core_distances[x],core_distances[i])
                dist = max(dists[x, i],core_distance_ptr[x],core_distance_ptr[i])
                if D[i] > dist:
                    D[i] = dist

                if D[i] < current_min:
                    y = i
                    current_min = D[i]

            Z[k, 0] = x
            Z[k, 1] = y
            Z[k, 2] = current_min
            x = y

    # Sort Z by cluster distances.
    order = np.argsort(Z_arr[:, 2], kind='mergesort')
    Z_arr = Z_arr[order]

    # Find correct cluster labels and compute cluster sizes inplace.
    label(Z_arr, n)

    return Z_arr


# Authors: Luis Scoccola
# License: 3-clause BSD

# from re import A
from turtle import back
import numpy as np
import warnings


import plotly.graph_objects as go
import plotly
from jupyter_dash import JupyterDash
import dash
from dash import dcc
from dash import html
from dash import DiskcacheManager
from dash.long_callback import DiskcacheLongCallbackManager

import pandas as pd
import json
import diskcache

PERSISTABLE_STDERR = "./persistable-stderr"
PERSISTABLE_DASH_CACHE = "./persistable-dash-cache"
# WARNINGS_GLOBAL = "test default"

X_START_FIRST_LINE = "x-start-first-line-"
Y_START_FIRST_LINE = "y-start-first-line-"
X_END_FIRST_LINE = "x-end-first-line-"
Y_END_FIRST_LINE = "y-end-first-line-"
X_START_SECOND_LINE = "x-start-second-line-"
Y_START_SECOND_LINE = "y-start-second-line-"
X_END_SECOND_LINE = "x-end-second-line-"
Y_END_SECOND_LINE = "y-end-second-line-"
DISPLAY_LINES_SELECTION = "display-lines-selection-"
ENDPOINT_SELECTION = "endpoint-selection-"
MIN_DIST_SCALE = "min-dist-scale-"
MAX_DIST_SCALE = "max-dist-scale-"
MIN_DENSITY_THRESHOLD = "min-density-threshold-"
MAX_DENSITY_THRESHOLD = "max-density-threshold-"
ENDPOINT_SELECTION_DIV = "endpoint-selection-div-"
PARAMETER_SELECTION_DIV = "parameter-selection-div-"
DISPLAY_PARAMETER_SELECTION = "display-parameter-selection-"

CFF_PLOT = "cff-plot-"
STORED_CCF = "stored-ccf-"
STORED_CCF_DRAWING = "stored-ccf-drawing-"
COMPUTE_CCF_BUTTON = "compute-ccf-button-"
INPUT_LOG_GRANULARITY_CCF = "input-log-granularity-ccf-"
INPUT_NUM_JOBS_CCF = "input-num-jobs-ccf-"
INPUT_MAX_COMPONENTS = "input-max-components-"
LOG = "log-"
WARNINGS_POLLING_INTERVAL = "warnings-polling-interval-"

PV_PLOT = "pv-plot-"
STORED_PV = "stored-pv-"
STORED_PV_DRAWING = "stored-pv-drawing-"
COMPUTE_PV_BUTTON = "compute-pv-button-"
INPUT_MAX_VINES = "input-max-vines-"
INPUT_PROM_VIN_SCALE = "input-prom-vin-scale-"
INPUT_LOG_GRANULARITY_PV = "input-log-granularity-pv-"
INPUT_NUM_JOBS_PV = "input-num-jobs-pv-"

VALUE = "value"
CLICKDATA = "clickData"
HIDDEN = "hidden"
DATA = "data"
N_CLICKS = "n_clicks"
DISABLED = "disabled"
FIGURE = "figure"
CHILDREN = "children"
N_INTERVALS = "n_intervals"

IN = "input"
ST = "state"


_funcs = []

def empty_figure():
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"fixedrange": True},
            yaxis={"fixedrange": True},
        )
    )
    fig.add_trace(go.Scatter(x=[], y=[]))
    fig.update_layout(template=None)
    fig.update_layout(autosize=True)
    fig.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig.update_layout(clickmode="none")
    return fig

def my_callback(
    inputs, outputs, prevent_initial_call=False, background=False, running=None
):
    def cs(l):
        return l[0] + l[1]

    def out_function(function):
        # FIX
        dash_inputs = [
            dash.Input(i, v) if t == IN else dash.State(i, v)
            for i, v, t in inputs
        ]
        dash_outputs = (
            [dash.Output(i, v) for i, v in outputs]
            if len(outputs) > 1
            else dash.Output(outputs[0][0], outputs[0][1])
        )

        if running is None:
            dash_running_outputs = None
        else:
            dash_running_outputs = [
                (dash.Output(i, v), value_start, value_end)
                for i, v, value_start, value_end in running
            ]

        def callback_function(*argv):
            d = {}
            for n, arg in enumerate(argv):
                d[cs(inputs[n])] = arg
            d = function(d)
            return (
                tuple(d[cs(o)] for o in outputs)
                if len(outputs) > 1
                else d[cs(outputs[0])]
            )

        global _funcs
        _funcs.append(callback_function)

        #print("value of background")
        #print(background)
        #print("callback function called with inputs:")
        #print(inputs)
        #print("___________________________________________________________________________")

        #if background:
        #    _app.long_callback(
        #        dash_outputs,
        #        dash_inputs,
        #        prevent_initial_call,
        #        running=dash_running_outputs,
        #    )(callback_function)
        #else :
        #    _app.callback(
        #        dash_outputs,
        #        dash_inputs,
        #        prevent_initial_call,
        #        running=dash_running_outputs,
        #        #background=background,
        #    )(callback_function)

        global _app
        _app.callback(
            dash_outputs,
            dash_inputs,
            prevent_initial_call,
            running=dash_running_outputs,
            background=background,
        )(callback_function)


        return function

    return out_function

# set temporary files
cache = diskcache.Cache(PERSISTABLE_DASH_CACHE)
#cache_2 = diskcache.Cache(PERSISTABLE_DASH_CACHE+"_2")
background_callback_manager = DiskcacheManager(cache)
#long_callback_manager = DiskcacheLongCallbackManager(cache)
#background_callback_manager_2 = DiskcacheManager(cache_2)
with open(PERSISTABLE_STDERR, "w") as file:
    file.close()

_app = dash.Dash(
    __name__,
    #long_callback_manager=long_callback_manager,
    background_callback_manager=background_callback_manager
)

@my_callback(
    [[DISPLAY_PARAMETER_SELECTION, VALUE, IN]],
    [[PARAMETER_SELECTION_DIV, HIDDEN]]
)
def toggle_parameter_selection_pv(d):
    if d[DISPLAY_PARAMETER_SELECTION + VALUE] == "on":
        d[PARAMETER_SELECTION_DIV + HIDDEN] = False
    else:
        d[PARAMETER_SELECTION_DIV + HIDDEN] = True
    return d

@my_callback(
    [[WARNINGS_POLLING_INTERVAL, N_INTERVALS, IN]], [[LOG, CHILDREN]], prevent_initial_call = False
)
def print_log(d):
    d = {}
    with open(PERSISTABLE_STDERR, "r") as file:
        d[LOG + CHILDREN] = file.read()
    return d

@my_callback(
    [
        [CFF_PLOT, CLICKDATA, IN],
        [DISPLAY_LINES_SELECTION, VALUE, ST],
        [ENDPOINT_SELECTION, VALUE, ST],
        [X_START_FIRST_LINE, VALUE, ST],
        [Y_START_FIRST_LINE, VALUE, ST],
        [X_END_FIRST_LINE, VALUE, ST],
        [Y_END_FIRST_LINE, VALUE, ST],
        [X_START_SECOND_LINE, VALUE, ST],
        [Y_START_SECOND_LINE, VALUE, ST],
        [X_END_SECOND_LINE, VALUE, ST],
        [Y_END_SECOND_LINE, VALUE, ST],
    ],
    [
        [X_START_FIRST_LINE, VALUE],
        [Y_START_FIRST_LINE, VALUE],
        [X_END_FIRST_LINE, VALUE],
        [Y_END_FIRST_LINE, VALUE],
        [X_START_SECOND_LINE, VALUE],
        [Y_START_SECOND_LINE, VALUE],
        [X_END_SECOND_LINE, VALUE],
        [Y_END_SECOND_LINE, VALUE],
    ],
    prevent_initial_call=True
)
def on_ccf_click(d):
    if d[DISPLAY_LINES_SELECTION + VALUE] == "on":
        new_x, new_y = (
            d[CFF_PLOT + CLICKDATA]["points"][0]["x"],
            d[CFF_PLOT + CLICKDATA]["points"][0]["y"],
        )
        if d[ENDPOINT_SELECTION + VALUE] == "1st line start":
            d[X_START_FIRST_LINE + VALUE] = new_x
            d[Y_START_FIRST_LINE + VALUE] = new_y
        elif d[ENDPOINT_SELECTION + VALUE] == "1st line end":
            d[X_END_FIRST_LINE + VALUE] = new_x
            d[Y_END_FIRST_LINE + VALUE] = new_y
        elif d[ENDPOINT_SELECTION + VALUE] == "2nd line start":
            d[X_START_SECOND_LINE + VALUE] = new_x
            d[Y_START_SECOND_LINE + VALUE] = new_y
        elif d[ENDPOINT_SELECTION + VALUE] == "2nd line end":
            d[X_END_SECOND_LINE + VALUE] = new_x
            d[Y_END_SECOND_LINE + VALUE] = new_y
    return d

@my_callback(
    [[DISPLAY_LINES_SELECTION, VALUE, IN]], [[ENDPOINT_SELECTION_DIV, HIDDEN]]
)
def toggle_parameter_selection_ccf(d):
    if d[DISPLAY_LINES_SELECTION + VALUE] == "on":
        d[ENDPOINT_SELECTION_DIV + HIDDEN] = False
    else:
        d[ENDPOINT_SELECTION_DIV + HIDDEN] = True
    return d

@my_callback(
    [[STORED_CCF, DATA, IN], [INPUT_MAX_COMPONENTS, VALUE, IN]],
    [[STORED_CCF_DRAWING, DATA]],
    prevent_initial_call=False
)
def draw_ccf(d):
    ccf = d[STORED_CCF + DATA]

    if ccf is None:
        d[STORED_CCF_DRAWING + DATA] = empty_figure()
        return d

    ccf = pd.read_json(ccf, orient="split")

    def df_to_plotly(df):
        return {
            "z": df.values.tolist(),
            "x": df.columns.tolist(),
            "y": df.index.tolist(),
        }

    fig = go.Figure(
        layout=go.Layout(
            xaxis_title="distance scale",
            yaxis_title="density threshold",
            xaxis={"fixedrange": True},
            yaxis={"fixedrange": True},
        ),
    )
    max_components = d[INPUT_MAX_COMPONENTS + VALUE]
    max_components = 0 if max_components is None else int(max_components)
    fig.add_trace(
        go.Heatmap(
            df_to_plotly(ccf),
            hovertemplate="<b># comp.: %{z:d}</b><br>x: %{x:.3e} <br>y: %{y:.3e} ",
            zmin=0,
            zmax=max_components,
            showscale=False,
            name="",
        )
    )
    fig.update_traces(colorscale="greys")

    fig.update_layout(showlegend=False)
    fig.update_layout(autosize=True)
    fig.update_xaxes(range=[min(ccf.columns), max(ccf.columns)])
    fig.update_yaxes(range=[min(ccf.index), max(ccf.index)])
    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
    fig.update_layout(clickmode="event+select")

    d[STORED_CCF_DRAWING + DATA] = plotly.io.to_json(fig)
    return d

@my_callback(
    [
        [STORED_CCF, DATA, ST],
        [STORED_CCF_DRAWING, DATA, IN],
        [MIN_DIST_SCALE, VALUE, IN],
        [MAX_DIST_SCALE, VALUE, IN],
        [MIN_DENSITY_THRESHOLD, VALUE, IN],
        [MAX_DENSITY_THRESHOLD, VALUE, IN],
        [DISPLAY_LINES_SELECTION, VALUE, IN],
        [X_START_FIRST_LINE, VALUE, IN],
        [Y_START_FIRST_LINE, VALUE, IN],
        [X_END_FIRST_LINE, VALUE, IN],
        [Y_END_FIRST_LINE, VALUE, IN],
        [X_START_SECOND_LINE, VALUE, IN],
        [Y_START_SECOND_LINE, VALUE, IN],
        [X_END_SECOND_LINE, VALUE, IN],
        [Y_END_SECOND_LINE, VALUE, IN],
        [ENDPOINT_SELECTION, VALUE, IN],
    ],
    [[CFF_PLOT, FIGURE]],
    prevent_initial_call=False
)
def draw_ccf_enclosing_box(d):
    ccf = d[STORED_CCF + DATA]

    if ccf is None:
        d[CFF_PLOT + FIGURE] = empty_figure()
        return d

    fig = plotly.io.from_json(d[STORED_CCF_DRAWING + DATA])

    ccf = pd.read_json(ccf, orient="split")

    def generate_red_box(xs, ys, text):
        return go.Scatter(
            x=xs,
            y=ys,
            fillcolor="rgba(255, 0, 0, 0.1)",
            fill="toself",
            mode="none",
            text=text,
            name="",
        )

    # draw left side of new enclosing box
    fig.add_trace(
        generate_red_box(
            [
                min(ccf.columns),
                d[MIN_DIST_SCALE + VALUE],
                d[MIN_DIST_SCALE + VALUE],
                min(ccf.columns),
            ],
            [min(ccf.index), min(ccf.index), max(ccf.index), max(ccf.index)],
            "Left side of new enclosing box",
        )
    )

    # draw right side of new enclosing box
    fig.add_trace(
        generate_red_box(
            [
                d[MAX_DIST_SCALE + VALUE],
                max(ccf.columns),
                max(ccf.columns),
                d[MAX_DIST_SCALE + VALUE],
            ],
            [min(ccf.index), min(ccf.index), max(ccf.index), max(ccf.index)],
            text="Right side of new enclosing box",
        )
    )

    # draw top side of new enclosing box
    fig.add_trace(
        generate_red_box(
            [
                d[MIN_DIST_SCALE + VALUE],
                d[MAX_DIST_SCALE + VALUE],
                d[MAX_DIST_SCALE + VALUE],
                d[MIN_DIST_SCALE + VALUE],
            ],
            [
                d[MAX_DENSITY_THRESHOLD + VALUE],
                d[MAX_DENSITY_THRESHOLD + VALUE],
                max(ccf.index),
                max(ccf.index),
            ],
            text="Top side of new enclosing box",
        )
    )

    # draw bottom side of new enclosing box
    fig.add_trace(
        generate_red_box(
            [
                d[MIN_DIST_SCALE + VALUE],
                d[MAX_DIST_SCALE + VALUE],
                d[MAX_DIST_SCALE + VALUE],
                d[MIN_DIST_SCALE + VALUE],
            ],
            [
                d[MIN_DENSITY_THRESHOLD + VALUE],
                d[MIN_DENSITY_THRESHOLD + VALUE],
                min(ccf.index),
                min(ccf.index),
            ],
            text="Bottom side of new enclosing box",
        )
    )

    if d[DISPLAY_LINES_SELECTION + VALUE] == "on":

        # draw polygon
        fig.add_trace(
            go.Scatter(
                x=[
                    d[X_START_FIRST_LINE + VALUE],
                    d[X_END_FIRST_LINE + VALUE],
                    d[X_END_SECOND_LINE + VALUE],
                    d[X_START_SECOND_LINE + VALUE],
                ],
                y=[
                    d[Y_START_FIRST_LINE + VALUE],
                    d[Y_END_FIRST_LINE + VALUE],
                    d[Y_END_SECOND_LINE + VALUE],
                    d[Y_START_SECOND_LINE + VALUE],
                ],
                fillcolor="rgba(0, 0, 255, 0.05)",
                fill="toself",
                mode="none",
                hoverinfo="skip",
                # text=text,
                name="",
            )
        )

        def generate_blue_line(xs, ys, text, different_marker=None):
            if different_marker == None:
                marker_styles = ["circle", "circle"]
            elif different_marker == 0:
                marker_styles = ["circle-open", "circle"]
            elif different_marker == 1:
                marker_styles = ["circle", "circle-open"]
            return go.Scatter(
                x=xs,
                y=ys,
                name=text + " line",
                text=[text + " line start", text + " line end"],
                marker=dict(size=15, color="blue"),
                # hoverinfo="name+text",
                marker_symbol=marker_styles,
                hoverinfo="skip",
                showlegend=False,
                mode="markers+lines+text",
                textposition="top center",
            )

        if d[ENDPOINT_SELECTION + VALUE] == "1st line start":
            first_line_endpoints = 0
            second_line_endpoints = None
        elif d[ENDPOINT_SELECTION + VALUE] == "1st line end":
            first_line_endpoints = 1
            second_line_endpoints = None
        elif d[ENDPOINT_SELECTION + VALUE] == "2nd line start":
            first_line_endpoints = None
            second_line_endpoints = 0
        elif d[ENDPOINT_SELECTION + VALUE] == "2nd line end":
            first_line_endpoints = None
            second_line_endpoints = 1

        fig.add_trace(
            generate_blue_line(
                [d[X_START_FIRST_LINE + VALUE], d[X_END_FIRST_LINE + VALUE]],
                [d[Y_START_FIRST_LINE + VALUE], d[Y_END_FIRST_LINE + VALUE]],
                "1st",
                first_line_endpoints,
            )
        )
        fig.add_trace(
            generate_blue_line(
                [d[X_START_SECOND_LINE + VALUE], d[X_END_SECOND_LINE + VALUE]],
                [d[Y_START_SECOND_LINE + VALUE], d[Y_END_SECOND_LINE + VALUE]],
                "2nd",
                second_line_endpoints,
            )
        )

    d[CFF_PLOT + FIGURE] = fig
    return d

@my_callback(
    [
        [STORED_PV, DATA, IN],
        [INPUT_MAX_VINES, VALUE, IN],
        [INPUT_PROM_VIN_SCALE, VALUE, IN],
    ],
    [[STORED_PV_DRAWING, DATA]],
    prevent_initial_call=False
)
def draw_pv(d):
    d[STORED_PV_DRAWING + DATA] = plotly.io.to_json(empty_figure())
    return d

@my_callback(
    [
        [STORED_PV_DRAWING, DATA, IN],
        [STORED_PV, DATA, ST],
    ],
    [
        [PV_PLOT, FIGURE]
    ]
)
def draw_pv_post(d):
    # pv,
    # pv_drawing,

    if d[STORED_PV + DATA] is None:
        d[PV_PLOT + FIGURE] = empty_figure()
        return d

    d[PV_PLOT + FIGURE] = plotly.io.from_json(d[STORED_PV_DRAWING + DATA])

    return d

@my_callback(
    [
        [COMPUTE_CCF_BUTTON, N_CLICKS, IN],
        [MIN_DENSITY_THRESHOLD, VALUE, ST],
        [MAX_DENSITY_THRESHOLD, VALUE, ST],
        [MIN_DIST_SCALE, VALUE, ST],
        [MAX_DIST_SCALE, VALUE, ST],
        [INPUT_LOG_GRANULARITY_CCF, VALUE, ST],
        [INPUT_NUM_JOBS_CCF, VALUE, ST],
    ],
    [
        [STORED_CCF, DATA],
    ],
    prevent_initial_call=True,
    background=False,
    running=[[COMPUTE_CCF_BUTTON, DISABLED, True, False]],
)
def compute_ccf(d):
    granularity = 2 ** d[INPUT_LOG_GRANULARITY_CCF + VALUE]
    num_jobs = int(d[INPUT_NUM_JOBS_CCF + VALUE])

    out = ""
    with warnings.catch_warnings(record=True) as w:
        # warnings.warn("test warning")
        print("1")
        global _persistable
        ss, ks, hf = _persistable.compute_hilbert_function(
            d[MIN_DENSITY_THRESHOLD + VALUE],
            d[MAX_DENSITY_THRESHOLD + VALUE],
            d[MIN_DIST_SCALE + VALUE],
            d[MAX_DIST_SCALE + VALUE],
            granularity,
            n_jobs=num_jobs,
        )
        print("2")
        for a in w:
            out += warnings.formatwarning(
                a.message, a.category, a.filename, a.lineno
            )
        with open(PERSISTABLE_STDERR, "w") as file:
            file.write(out)

    d[STORED_CCF + DATA] = pd.DataFrame(
        hf, index=ks[:-1], columns=ss[:-1]
    ).to_json(date_format="iso", orient="split")

    return d

@my_callback(
    [
        [COMPUTE_PV_BUTTON, N_CLICKS, IN],
        [X_START_FIRST_LINE, VALUE, ST],
        [Y_START_FIRST_LINE, VALUE, ST],
        [X_END_FIRST_LINE, VALUE, ST],
        [Y_END_FIRST_LINE, VALUE, ST],
        [X_START_SECOND_LINE, VALUE, ST],
        [Y_START_SECOND_LINE, VALUE, ST],
        [X_END_SECOND_LINE, VALUE, ST],
        [Y_END_SECOND_LINE, VALUE, ST],
        [INPUT_LOG_GRANULARITY_PV, VALUE, ST],
        [INPUT_NUM_JOBS_PV, VALUE, ST],
    ],
    [
        [STORED_PV, DATA]
    ],
    prevent_initial_call=True,
    background=False,
    running=[[COMPUTE_PV_BUTTON, DISABLED, True, False]],
)
def compute_pv(d):

    granularity = 2**d[INPUT_LOG_GRANULARITY_PV+VALUE]
    num_jobs = int(d[INPUT_NUM_JOBS_PV+VALUE])

    out = ""
    with warnings.catch_warnings(record=True) as w:
        global _persistable
        pv = _persistable.compute_prominence_vineyard(
            [
                [d[X_START_FIRST_LINE+VALUE], d[Y_START_FIRST_LINE+VALUE]],
                [d[X_END_FIRST_LINE+VALUE], d[Y_END_FIRST_LINE+VALUE]],
            ],
            [
                [d[X_START_SECOND_LINE+VALUE], d[Y_START_SECOND_LINE+VALUE]],
                [d[X_END_SECOND_LINE+VALUE], d[Y_END_SECOND_LINE+VALUE]],
            ],
            n_parameters=granularity,
            n_jobs=num_jobs,
        )
        for a in w:
            out += warnings.formatwarning(
                a.message, a.category, a.filename, a.lineno
            )
        with open(PERSISTABLE_STDERR, "w") as file:
            file.write(out)

    d[STORED_PV+DATA] = json.dumps(pv.__dict__)

    return d


def PersistableInteractive(persistable, debug=False):

    global _app
    global _persistable
    _persistable = persistable

    default_min_k = 0
    default_max_k = _persistable._maxk
    default_k_step = default_max_k / 100
    default_min_s = _persistable._connection_radius / 5
    default_max_s = _persistable._connection_radius * 2
    default_s_step = (default_max_s - default_min_s) / 100
    default_log_granularity = 6
    default_num_jobs = 4
    default_max_dim = 15
    default_max_vines = 20
    min_granularity = 4
    max_granularity = 8
    defr = 6
    default_x_start_first_line = (default_min_s + default_max_s) * (1 / defr)
    default_y_start_first_line = (default_min_k + default_max_k) * (1 / 2)
    default_x_end_first_line = (default_max_s + default_min_s) * (1 / 2)
    default_y_end_first_line = (default_min_k + default_max_k) * (1 / defr)
    default_x_start_second_line = (default_min_s + default_max_s) * (1 / 2)
    default_y_start_second_line = (default_min_k + default_max_k) * (
        (defr - 1) / defr
    )
    default_x_end_second_line = (default_max_s + default_min_s) * (
        (defr - 1) / defr
    )
    default_y_end_second_line = (default_min_k + default_max_k) * (1 / 2)


    _app.layout = html.Div(
        children=[
            # contains the component counting function as a pandas dataframe
            dcc.Store(id=STORED_CCF),
            # contains the basic component counting function plot as a plotly figure
            dcc.Store(id=STORED_CCF_DRAWING),
            # contains the vineyard as a vineyard object
            dcc.Store(id=STORED_PV),
            # contains the basic prominence vineyard plot as a plotly figure
            dcc.Store(id=STORED_PV_DRAWING),
            dcc.Interval(
                id=WARNINGS_POLLING_INTERVAL,
                interval=(1 / 2) * 1000,
                n_intervals=0,
            ),
            html.H1("Interactive parameter selection for Persistable"),
            html.Details(
                [
                    html.Summary("Quick help"),
                    dcc.Markdown(
                        """
                    - When setting a field, press enter to TODO
                    - Check the log, below, for warnings.
                    - The app takes a second or so to update the graphical interface after an interaction.
                    - Computing the component counting function and prominence vineyard can take a while, depending on the size and dimensionality of the dataset as well as other factors.
                    - TODO
                    """
                    ),
                ]
            ),
            html.Div(
                className="grid",
                children=[
                    html.Div(
                        children=[
                            html.H2("Component Counting Function"),
                            html.Div(
                                className="parameters",
                                children=[
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="density threshold min/max",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=MIN_DENSITY_THRESHOLD,
                                                type="number",
                                                value=default_min_k,
                                                min=0,
                                                debounce=True,
                                                step=default_k_step,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=MAX_DENSITY_THRESHOLD,
                                                type="number",
                                                value=default_max_k,
                                                min=0,
                                                debounce=True,
                                                step=default_k_step,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="distance scale min/max",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=MIN_DIST_SCALE,
                                                type="number",
                                                value=default_min_s,
                                                min=0,
                                                debounce=True,
                                                step=default_s_step,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=MAX_DIST_SCALE,
                                                type="number",
                                                value=default_max_s,
                                                min=0,
                                                debounce=True,
                                                step=default_s_step,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="number of cores",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=INPUT_NUM_JOBS_CCF,
                                                type="number",
                                                value=default_num_jobs,
                                                min=1,
                                                step=1,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="granularity comp count func",
                                            ),
                                            dcc.Slider(
                                                min_granularity,
                                                max_granularity,
                                                step=None,
                                                marks={
                                                    i: str(2**i)
                                                    for i in range(
                                                        1, max_granularity + 1
                                                    )
                                                },
                                                value=default_log_granularity,
                                                id=INPUT_LOG_GRANULARITY_CCF,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                            ),
                                            html.Button(
                                                "(re)compute component counting function",
                                                id=COMPUTE_CCF_BUTTON,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="max connected components",
                                            ),
                                            dcc.Input(
                                                id=INPUT_MAX_COMPONENTS,
                                                type="number",
                                                value=default_max_dim,
                                                min=1,
                                                className=VALUE,
                                                step=1,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="lines selection",
                                            ),
                                            dcc.RadioItems(
                                                ["on", "off"],
                                                "off",
                                                id=DISPLAY_LINES_SELECTION,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        id=ENDPOINT_SELECTION_DIV,
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="endpoint selection",
                                            ),
                                            dcc.RadioItems(
                                                [
                                                    "1st line start",
                                                    "1st line end",
                                                    "2nd line start",
                                                    "2nd line end",
                                                ],
                                                "1st line start",
                                                id=ENDPOINT_SELECTION,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ]
                    ),
                    html.Div(
                        children=[
                            html.H2("Prominence Vineyard"),
                            html.Div(
                                className="parameters",
                                children=[
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="first line start x/y",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=X_START_FIRST_LINE,
                                                type="number",
                                                value=default_x_start_first_line,
                                                min=0,
                                                step=default_s_step,
                                                debounce=True,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=Y_START_FIRST_LINE,
                                                type="number",
                                                value=default_y_start_first_line,
                                                min=0,
                                                step=default_k_step,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="first line end x/y",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=X_END_FIRST_LINE,
                                                type="number",
                                                value=default_x_end_first_line,
                                                min=0,
                                                step=default_s_step,
                                                debounce=True,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=Y_END_FIRST_LINE,
                                                type="number",
                                                value=default_y_end_first_line,
                                                min=0,
                                                step=default_k_step,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="second line start x/y",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=X_START_SECOND_LINE,
                                                type="number",
                                                value=default_x_start_second_line,
                                                min=0,
                                                step=default_s_step,
                                                debounce=True,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=Y_START_SECOND_LINE,
                                                type="number",
                                                value=default_y_start_second_line,
                                                min=0,
                                                step=default_k_step,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-double",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="second line end x/y",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=X_END_SECOND_LINE,
                                                type="number",
                                                value=default_x_end_second_line,
                                                min=0,
                                                step=default_s_step,
                                                debounce=True,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=Y_END_SECOND_LINE,
                                                type="number",
                                                value=default_y_end_second_line,
                                                min=0,
                                                step=default_k_step,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="number of cores",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id=INPUT_NUM_JOBS_PV,
                                                type="number",
                                                value=default_num_jobs,
                                                min=1,
                                                step=1,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="granularity vineyard",
                                            ),
                                            dcc.Slider(
                                                min_granularity,
                                                max_granularity,
                                                step=None,
                                                marks={
                                                    i: str(2**i)
                                                    for i in range(
                                                        1, max_granularity + 1
                                                    )
                                                },
                                                value=default_log_granularity - 1,
                                                id=INPUT_LOG_GRANULARITY_PV,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                            ),
                                            html.Button(
                                                "compute prominence vineyard",
                                                id=COMPUTE_PV_BUTTON,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="max number vines",
                                            ),
                                            dcc.Input(
                                                id=INPUT_MAX_VINES,
                                                type="number",
                                                value=default_max_vines,
                                                min=1,
                                                className=VALUE,
                                                step=1,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="prominence scale",
                                            ),
                                            dcc.RadioItems(
                                                ["linear", "logarithmic"],
                                                "logarithmic",
                                                id=INPUT_PROM_VIN_SCALE,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-single",
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="parameter selection",
                                            ),
                                            dcc.RadioItems(
                                                ["on", "off"],
                                                "off",
                                                id=DISPLAY_PARAMETER_SELECTION,
                                                className=VALUE,
                                            ),
                                        ],
                                    ),
                                    html.Div(
                                        className="parameter-double",
                                        id=PARAMETER_SELECTION_DIV,
                                        children=[
                                            html.Span(
                                                className="name",
                                                children="line/gap",
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id="input-line",
                                                type="number",
                                                value=1,
                                                min=1,
                                                debounce=True,
                                            ),
                                            dcc.Input(
                                                className=VALUE,
                                                id="gap",
                                                type="number",
                                                value=1,
                                                min=1,
                                                debounce=True,
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ]
                    ),
                    dcc.Graph(
                        id=CFF_PLOT,
                        figure=empty_figure(),
                        config={
                            "responsive": True,
                            "displayModeBar": False,
                            "modeBarButtonsToRemove": [
                                "toImage",
                                "pan",
                                "zoomIn",
                                "zoomOut",
                                "resetScale",
                                "lasso",
                                "select",
                            ],
                            "displaylogo": False,
                        },
                    ),
                    dcc.Graph(
                        id=PV_PLOT,
                        figure=empty_figure(),
                        config={
                            "responsive": True,
                            "displayModeBar": False,
                            "modeBarButtonsToRemove": [
                                "toImage",
                                "pan",
                                "zoomIn",
                                "zoomOut",
                                "resetScale",
                                "lasso",
                                "select",
                            ],
                            "displaylogo": False,
                        },
                    ),
                ],
            ),
            html.Details(
                [
                    html.Summary("Warnings"),
                    html.Pre(
                        id=LOG,
                        style={
                            "border": "thin lightgrey solid",
                            "overflowX": "scroll",
                        },
                    ),
                ],
                open=True,
            ),
        ],
    )

    _app.run_server(debug=debug)



####        def compute_ccf(
####            n_clicks,
####            min_k,
####            max_k,
####            min_s,
####            max_s,
####            log_granularity,
####            num_jobs,
####        ):
####            print("compute_ccf was run")
####            granularity = 2**log_granularity
####            num_jobs = int(num_jobs)
####    
####            out = ""
####            with warnings.catch_warnings(record=True) as w:
####                print("compute_ccf entered the warning part")
####                # Cause all warnings to always be triggered.
####                # warnings.simplefilter("always")
####                # warnings.warn("test warning")
####                ss, ks, hf = _persistable.compute_hilbert_function(
####                    min_k,
####                    max_k,
####                    min_s,
####                    max_s,
####                    granularity,
####                    n_jobs=num_jobs,
####                )
####                print("compute_ccf did the difficult part")
####                for a in w:
####                    out += warnings.formatwarning(
####                        a.message, a.category, a.filename, a.lineno
####                    )
####                with open(PERSISTABLE_STDERR, "w") as file:
####                    file.write(out)
####    
####            return pd.DataFrame(hf, index=ks[:-1], columns=ss[:-1]).to_json(
####                date_format="iso", orient="split"
####            )
####    
####        def compute_pv(
####            n_clicks,
####            x_start_first_line,
####            y_start_first_line,
####            x_end_first_line,
####            y_end_first_line,
####            x_start_second_line,
####            y_start_second_line,
####            x_end_second_line,
####            y_end_second_line,
####            log_granularity,
####            num_jobs,
####        ):
####            granularity = 2**log_granularity
####            num_jobs = int(num_jobs)
####    
####            out = ""
####            with warnings.catch_warnings(record=True) as w:
####                # Cause all warnings to always be triggered.
####                # warnings.simplefilter("always")
####                print("compute_pv entered the warning part")
####                pv = _persistable.compute_prominence_vineyard(
####                    [
####                        [x_start_first_line, y_start_first_line],
####                        [x_end_first_line, y_end_first_line],
####                    ],
####                    [
####                        [x_start_second_line, y_start_second_line],
####                        [x_end_second_line, y_end_second_line],
####                    ],
####                    n_parameters=granularity,
####                    n_jobs=num_jobs,
####                )
####                print("compute_pv did the difficult part")
####                for a in w:
####                    out += warnings.formatwarning(
####                        a.message, a.category, a.filename, a.lineno
####                    )
####                with open(PERSISTABLE_STDERR, "w") as file:
####                    file.write(out)
####    
####            return json.dumps(pv.__dict__)
####
####        _app.callback(
####            dash.Output(STORED_CCF, DATA),
####            [
####                dash.Input(COMPUTE_CCF_BUTTON, N_CLICKS),
####                dash.State(MIN_DENSITY_THRESHOLD, VALUE),
####                dash.State(MAX_DENSITY_THRESHOLD, VALUE),
####                dash.State(MIN_DIST_SCALE, VALUE),
####                dash.State(MAX_DIST_SCALE, VALUE),
####                dash.State(INPUT_LOG_GRANULARITY_CCF, VALUE),
####                dash.State(INPUT_NUM_JOBS_CCF, VALUE),
####            ],
####            True,
####            running=[(dash.Output(COMPUTE_CCF_BUTTON, DISABLED), True, False)],
####        )(compute_ccf)
####
####        _app.callback(
####            dash.Output(STORED_PV, DATA),
####            [
####                dash.Input(COMPUTE_PV_BUTTON, N_CLICKS),
####                dash.State(X_START_FIRST_LINE, VALUE),
####                dash.State(Y_START_FIRST_LINE, VALUE),
####                dash.State(X_END_FIRST_LINE, VALUE),
####                dash.State(Y_END_FIRST_LINE, VALUE),
####                dash.State(X_START_SECOND_LINE, VALUE),
####                dash.State(Y_START_SECOND_LINE, VALUE),
####                dash.State(X_END_SECOND_LINE, VALUE),
####                dash.State(Y_END_SECOND_LINE, VALUE),
####                dash.State(INPUT_LOG_GRANULARITY_CCF, VALUE),
####                dash.State(INPUT_NUM_JOBS_CCF, VALUE),
####            ],
####            True,
####            running=[(dash.Output(COMPUTE_PV_BUTTON, DISABLED), True, False)],
####        )(compute_pv)



#        def compute_ccf(d):
#            granularity = 2 ** d[INPUT_LOG_GRANULARITY_CCF + VALUE]
#            num_jobs = int(d[INPUT_NUM_JOBS_CCF + VALUE])
#
#            out = ""
#            with warnings.catch_warnings(record=True) as w:
#                # warnings.warn("test warning")
#                ss, ks, hf = _persistable.compute_hilbert_function(
#                    d[MIN_DENSITY_THRESHOLD + VALUE],
#                    d[MAX_DENSITY_THRESHOLD + VALUE],
#                    d[MIN_DIST_SCALE + VALUE],
#                    d[MAX_DIST_SCALE + VALUE],
#                    granularity,
#                    n_jobs=num_jobs,
#                )
#                for a in w:
#                    out += warnings.formatwarning(
#                        a.message, a.category, a.filename, a.lineno
#                    )
#                with open(PERSISTABLE_STDERR, "w") as file:
#                    file.write(out)
#
#            d[STORED_CCF + DATA] = pd.DataFrame(
#                hf, index=ks[:-1], columns=ss[:-1]
#            ).to_json(date_format="iso", orient="split")
#            return d
#
#        my_callback(
#            [
#                [COMPUTE_CCF_BUTTON, N_CLICKS, IN],
#                [MIN_DENSITY_THRESHOLD, VALUE, ST],
#                [MAX_DENSITY_THRESHOLD, VALUE, ST],
#                [MIN_DIST_SCALE, VALUE, ST],
#                [MAX_DIST_SCALE, VALUE, ST],
#                [INPUT_LOG_GRANULARITY_CCF, VALUE, ST],
#                [INPUT_NUM_JOBS_CCF, VALUE, ST],
#            ],
#            [
#                [STORED_CCF, DATA],
#            ],
#            prevent_initial_call=True,
#            background=True,
#            running=[[COMPUTE_CCF_BUTTON, DISABLED, True, False]],
#            manager=background_callback_manager
#        )(compute_ccf)
#
#        def compute_pv(d):
#            granularity = 2**d[INPUT_LOG_GRANULARITY_PV+VALUE]
#            num_jobs = int(d[INPUT_NUM_JOBS_PV+VALUE])
#
#            out = ""
#            with warnings.catch_warnings(record=True) as w:
#                pv = _persistable.compute_prominence_vineyard(
#                    [
#                        [d[X_START_FIRST_LINE+VALUE], d[Y_START_FIRST_LINE+VALUE]],
#                        [d[X_END_FIRST_LINE+VALUE], d[Y_END_FIRST_LINE+VALUE]],
#                    ],
#                    [
#                        [d[X_START_SECOND_LINE+VALUE], d[Y_START_SECOND_LINE+VALUE]],
#                        [d[X_END_SECOND_LINE+VALUE], d[Y_END_SECOND_LINE+VALUE]],
#                    ],
#                    n_parameters=granularity,
#                    n_jobs=num_jobs,
#                )
#                for a in w:
#                    out += warnings.formatwarning(
#                        a.message, a.category, a.filename, a.lineno
#                    )
#                with open(PERSISTABLE_STDERR, "w") as file:
#                    file.write(out)
#
#            d[STORED_PV+DATA] = json.dumps(pv.__dict__)
#
#            return d
#
#        my_callback(
#            [
#                [COMPUTE_PV_BUTTON, N_CLICKS, IN],
#                [X_START_FIRST_LINE, VALUE, ST],
#                [Y_START_FIRST_LINE, VALUE, ST],
#                [X_END_FIRST_LINE, VALUE, ST],
#                [Y_END_FIRST_LINE, VALUE, ST],
#                [X_START_SECOND_LINE, VALUE, ST],
#                [Y_START_SECOND_LINE, VALUE, ST],
#                [X_END_SECOND_LINE, VALUE, ST],
#                [Y_END_SECOND_LINE, VALUE, ST],
#                [INPUT_LOG_GRANULARITY_PV, VALUE, ST],
#                [INPUT_NUM_JOBS_PV, VALUE, ST],
#            ],
#            [
#                [STORED_PV, DATA]
#            ],
#            prevent_initial_call=True,
#            background=True,
#            running=[[COMPUTE_PV_BUTTON, DISABLED, True, False]],
#            manager=background_callback_manager_2
#        )(compute_pv)




#    def plot_prominence_vineyard(
#        self,
#        vineyard,
#        interpolate=True,
#        areas=True,
#        points=False,
#        log_prominence=True,
#        colormap="viridis",
#    ):
#        ax = self._vineyard_ax
#
#        # TODO: abstract this
#        ax.clear()
#        self._gaps = []
#        self._gap_numbers = []
#        self._lines = []
#        self._line_index = []
#
#        times = vineyard._parameter_indices
#        vines = vineyard._vineyard_to_vines()
#        num_vines = min(len(vines), vineyard._firstn)
#
#        ax.set_title("prominence vineyard")
#
#        # TODO: warn that vineyard is empty
#        if num_vines == 0:
#            return
#
#        cmap = cm.get_cmap(colormap)
#        colors = list(cmap(np.linspace(0, 1, num_vines)[::-1]))
#        last = colors[-1]
#        colors.extend([last for _ in range(num_vines - vineyard._firstn)])
#        if areas:
#            for i in range(len(vines) - 1):
#                artist = ax.fill_between(
#                    times, vines[i][1], vines[i + 1][1], color=colors[i]
#                )
#                self._add_gap_prominence_vineyard(artist, i + 1)
#            artist = ax.fill_between(
#                times, vines[len(vines) - 1][1], 0, color=colors[len(vines) - 1]
#            )
#            self._add_gap_prominence_vineyard(artist, len(vines))
#        for i, tv in enumerate(vines):
#            times, vine = tv
#            for vine_part, time_part in vineyard._vine_parts(vine):
#                if interpolate:
#                    artist = ax.plot(time_part, vine_part, c="black")
#                if points:
#                    artist = ax.plot(time_part, vine_part, "o", c="black")
#                self._vineyard_values.extend(vine_part)
#        ymax = max(self._vineyard_values)
#        for t in times:
#            artist = ax.vlines(x=t, ymin=0, ymax=ymax, color="black", alpha=0.1)
#            self._add_line_prominence_vineyard(artist, t)
#        ax.set_xticks([])
#        ax.set_xlabel("parameter")
#        if log_prominence:
#            ax.set_ylabel("log-prominence")
#            ax.set_yscale(LOG)
#        else:
#            ax.set_ylabel("prominence")
#        values = np.array(self._vineyard_values)
#
#        ax.set_ylim([np.quantile(values[values > 0], 0.05), max(values)])
#        ax.figure.canvas.draw_idle()
#        ax.figure.canvas.flush_events()
#
#    def _vineyard_on_parameter_selection(self, event):
#        ax = self._vineyard_ax
#        if event.inaxes != ax:
#            return
#
#        if event.button == 1:
#            # info = ""
#
#            # gaps
#            gap = None
#            aas = []
#            for aa, artist in enumerate(self._gaps):
#                cont, _ = artist.contains(event)
#                if not cont:
#                    continue
#                aas.append(aa)
#            if len(aas) > 0:
#                # aa = aas[-1]
#                gap = aas[-1]
#                # lbl = self._gap_numbers[aa]
#                # info += "gap: " + str(lbl) + ";    "
#
#            # lines
#            line_index = None
#            aas = []
#            for aa, artist in enumerate(self._lines):
#                cont, _ = artist.contains(event)
#                if not cont:
#                    continue
#                aas.append(aa)
#            if len(aas) > 0:
#                # aa = aas[-1]
#                line_index = aas[-1]
#                # lbl = self._line_index[aa]
#                # info += "line: " + str(lbl) + ";    "
#
#            if gap is not None and line_index is not None:
#                parameters, n_clusters = self._update_line_parameters(
#                    gap + 1, line_index
#                )
#                if self._vineyard_current_points_plotted_on is not None:
#                    self._vineyard_current_points_plotted_on.remove()
#                self._vineyard_current_points_plotted_on = ax.scatter(
#                    [event.xdata], [event.ydata], c="blue", s=40
#                )
#
#                info = "Parameter ({:.2f}, {:.2f}) -> ({:.2f}, {:.2f}), with n_clusters = {:d} selected.".format(
#                    parameters[0][0],
#                    parameters[0][1],
#                    parameters[1][0],
#                    parameters[1][1],
#                    n_clusters,
#                )
#                ax.format_coord = lambda x, y: info
#
#                ax.figure.canvas.draw_idle()
#                ax.figure.canvas.flush_events()
#
#    def _draw_on_hilbert(self, vineyard_parameters):
#        ax = self._hilbert_ax
#        points = np.array(list(vineyard_parameters.values()))
#
#        self._hilbert_current_points_plotted_on = ax.scatter(
#            points[:, 0], points[:, 1], c="blue", s=10
#        )
#        if len(points) >= 2:
#            self._hilbert_current_lines_plotted_on.append(
#                ax.plot(
#                    [points[0, 0], points[1, 0]],
#                    [points[0, 1], points[1, 1]],
#                    c="blue",
#                    linewidth=1,
#                )
#            )
#        if len(points) >= 4:
#            self._hilbert_current_lines_plotted_on.append(
#                ax.plot(
#                    [points[2, 0], points[3, 0]],
#                    [points[2, 1], points[3, 1]],
#                    c="blue",
#                    linewidth=1,
#                )
#            )
#            polygon = Polygon(
#                [points[0], points[1], points[3], points[2]],
#                True,
#                color="red",
#                alpha=0.1,
#            )
#            ax.add_patch(polygon)
#            self._hilbert_current_polygon_plotted_on = polygon
#        if len(points) >= 4:
#            info = "Prominence vineyard with ({:.2f}, {:.2f}) -> ({:.2f}, {:.2f}) to ({:.2f}, {:.2f}) -> ({:.2f}, {:.2f}) selected.".format(
#                points[0, 0],
#                points[0, 1],
#                points[1, 0],
#                points[1, 1],
#                points[2, 0],
#                points[2, 1],
#                points[3, 0],
#                points[3, 1],
#            )
#            ax.format_coord = lambda x, y: info
#
#        ax.figure.canvas.draw_idle()
#        ax.figure.canvas.flush_events()
#
#    def _hilbert_on_parameter_selection(self, event):
#        ax = self._hilbert_ax
#        if event.inaxes != ax:
#            return
#        if event.button == 1:
#            vineyard_parameters = self._update_vineyard_parameter_bounds(
#                [event.xdata, event.ydata]
#            )
#            self._clear_hilbert_parameters()
#            self._draw_on_hilbert(vineyard_parameters)
#
#    def _hilbert_on_clear_parameter(self, event):
#        _ = self._clear_vineyard_parameter_bounds()
#        self._clear_hilbert_parameters()
#
#    def _add_gap_prominence_vineyard(self, artist, number):
#
#        if isinstance(artist, list):
#            assert len(artist) == 1
#            artist = artist[0]
#
#        self._gaps += [artist]
#        self._gap_numbers += [number]
#
#    def _add_line_prominence_vineyard(self, artist, number):
#
#        if isinstance(artist, list):
#            assert len(artist) == 1
#            artist = artist[0]
#
#        self._lines += [artist]
#        self._line_index += [number]



# Authors: Luis Scoccola
# License: 3-clause BSD

import numpy as np
import warnings
import traceback
import plotly.graph_objects as go
import plotly
from plotly.colors import sample_colorscale
import json
#import diskcache
import dash
from dash import dcc, html, DiskcacheManager, ctx
from dash.exceptions import PreventUpdate
from jupyter_dash import JupyterDash
import click
import socket
from ._vineyard import Vineyard

import threading


# monkeypatch the hashing function of dash, so that
# we can use a decorator to register callbacks
from dash.long_callback.managers import BaseLongCallbackManager

import uuid


def monkeypatched_hash_function(fn):
    return uuid.uuid4()


BaseLongCallbackManager.hash_function = monkeypatched_hash_function
###


X_START_FIRST_LINE = "x-start-first-line-"
Y_START_FIRST_LINE = "y-start-first-line-"
X_END_FIRST_LINE = "x-end-first-line-"
Y_END_FIRST_LINE = "y-end-first-line-"
X_START_SECOND_LINE = "x-start-second-line-"
Y_START_SECOND_LINE = "y-start-second-line-"
X_END_SECOND_LINE = "x-end-second-line-"
Y_END_SECOND_LINE = "y-end-second-line-"
CCF_PLOT = "ccf-plot-"
DISPLAY_LINES_SELECTION = "display-lines-selection-"
ENDPOINT_SELECTION = "endpoint-selection-"
STORED_CCF = "stored-ccf-"
STORED_CCF_DRAWING = "stored-ccf-drawing-"
MIN_DIST_SCALE = "min-dist-scale-"
MAX_DIST_SCALE = "max-dist-scale-"
MIN_DENSITY_THRESHOLD = "min-density-threshold-"
MAX_DENSITY_THRESHOLD = "max-density-threshold-"
ENDPOINT_SELECTION_DIV = "endpoint-selection-div-"
PARAMETER_SELECTION_DIV = "parameter-selection-div-"
DISPLAY_PARAMETER_SELECTION = "display-parameter-selection-"
COMPUTE_CCF_BUTTON = "compute-ccf-button-"
STOP_COMPUTE_CCF_BUTTON = "stop-compute-ccf-button-"
INPUT_GRANULARITY_CCF = "input-granularity-ccf-"
INPUT_NUM_JOBS_CCF = "input-num-jobs-ccf-"
INPUT_MAX_COMPONENTS = "input-max-components-"
CCF_PLOT_CONTROLS_DIV = "ccf-plot-controls-div-"
CCF_DETAILS = "ccf-details-"
PV_DETAILS = "pv-details-"
PV_PLOT_CONTROLS_DIV = "pv-plot-controls-div-"
LOG = "log-"
LOG_DIV = "log-div-"
STORED_PV = "stored-pv-"
INPUT_MAX_VINES = "input-max-vines-"
INPUT_PROM_VIN_SCALE = "input-prom-vin-scale-"
COMPUTE_PV_BUTTON = "compute-pv-button-"
STOP_COMPUTE_PV_BUTTON = "stop-compute-pv-button-"
PV_PLOT = "pv-plot-"
STORED_PV_DRAWING = "stored-pv-drawing-"
INPUT_GRANULARITY_PV = "input-granularity-pv-"
INPUT_NUM_JOBS_PV = "input-num-jobs-pv-"
INPUT_LINE = "input-line-"
INPUT_GAP = "gap-"
EXPORT_PARAMETERS_BUTTON = "export-parameters-"
FIXED_PARAMETERS = "fixed-parameters-"
STORED_PD = "stored-pd-"

STORED_CCF_COMPUTATION_WARNINGS = "stored-ccf-computation-warnings-"
STORED_PV_COMPUTATION_WARNINGS = "stored-pv-computation-warnings-"
# STORED_PD_COMPUTATION_WARNINGS = "stored-pd-computation-warnings-"

EXPORTED_PARAMETER = "exported-parameter-"

SHUTDOWN_BUTTON = "shutdown-button-"

VALUE = "value"
CLICKDATA = "clickData"
HIDDEN = "hidden"
DATA = "data"
N_CLICKS = "n_clicks"
DISABLED = "disabled"
FIGURE = "figure"
CHILDREN = "children"
N_INTERVALS = "n_intervals"

IN = "input"
ST = "state"


def empty_figure():
    fig = go.Figure(
        layout=go.Layout(
            xaxis={"fixedrange": True},
            yaxis={"fixedrange": True},
        )
    )
    fig.add_trace(go.Scatter(x=[], y=[]))
    fig.update_layout(template=None)
    fig.update_layout(autosize=True)
    fig.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig.update_layout(clickmode="none")
    return fig


class PersistableInteractive:
    """A graphical user interface for doing parameter selection for Persistable.

    persistable: Persistable
        Persistable instance with which to interact with the user interface.

    """

    def __init__(self, persistable):
        self._persistable = persistable
        self._app = None
        self._debug = False
        self._parameters_sem = threading.Semaphore()
        self._parameters = None


    def start_UI(self, port=8050, debug=False, inline=False):
        """Serves the GUI with a given persistable instance.

        port: int, optional, default is 8050
            Integer representing which port of localhost to try use to run the GUI.
            If port is not available, we look for one that is available, starting
            from the given one.

        debug: bool, optional, default is False
            Whether to run Dash in debug mode.

        inline: bool, optional, default is False
            Boolean representing whether or not to display the GUI inside
            the Jupyter notebook. Only works if using Persistable in a Jupyter notebook.

        return: int
            If not run in inline mode, returns the port of localhost used to serve the UI.

        """
        if debug:
            self._debug = debug
        max_port = 65535
        for possible_port in range(port, max_port+1):
            # check if port is in use
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                in_use = s.connect_ex(("localhost", int(possible_port))) == 0
            if in_use:
                continue
            else:
                port = possible_port
                break
        if possible_port == max_port:
            raise Exception(
                "All ports are already in use. Cannot start the GUI."
            )

        background_callback_manager = DiskcacheManager()

        def suppress_warnings(app):
            import logging
            app.logger.setLevel(logging.WARNING)
            logging.getLogger("werkzeug").setLevel(logging.ERROR)
            def secho(text, file=None, nl=None, err=None, color=None, **styles):
                pass
            def echo(text, file=None, nl=None, err=None, color=None, **styles):
                pass
            click.echo = echo
            click.secho = secho

        if inline == True:

            self._app = JupyterDash(
                __name__,
                background_callback_manager=background_callback_manager,
                update_title = "Persistable is computing..."
            )
            self._layout_gui()
            self._register_callbacks(self._persistable, self._debug)

            if not debug:
                suppress_warnings(self._app)

            self._app.run_server(port=port, mode="inline", debug=debug)
        else:
            self._app = dash.Dash(
                __name__,
                background_callback_manager=background_callback_manager,
                update_title = "Persistable is computing..."
            )
            self._layout_gui()
            self._register_callbacks(self._persistable, self._debug)

            if not debug:
                suppress_warnings(self._app)

            def run():
                self._app.run_server(port=port, debug=debug, use_reloader=False)

            self._thread = threading.Thread(target=run)
            self._thread.daemon = True
            self._thread.start()

            return port


    def cluster(self, **kwargs):
        """Clusters the dataset with which the Persistable instance that was
        passed through ``run_with`` was initialized.

        ``**kwargs``:
            Passed to ``Persistable.cluster``.

        returns:
            A numpy array of length the number of points in the dataset containing
            integers from -1 to the number of clusters minus 1, representing the
            labels of the final clustering. The label -1 represents noise points,
            i.e., points deemed not to belong to any cluster by the algorithm.

        """
        params = self._chosen_parameters()
        if params == None:
            raise ValueError(
                "No parameters where chosen. Please use the graphical user interface to choose parameters."
            )
        else:
            return self._persistable.cluster(**params, **kwargs)


    def _chosen_parameters(self):
        self._parameters_sem.acquire()
        params = self._parameters
        self._parameters_sem.release()
        return params


    def _layout_gui(self):
        default_min_k = 0
        end = self._persistable._find_end()
        default_max_k = end[1]
        #default_k_step = default_max_k / 100
        default_min_s = 0
        default_max_s = end[0]
        #default_s_step = (default_max_s - default_min_s) / 100
        default_granularity_ccf = 100
        default_granularity_pv = 40
        default_num_jobs = 4
        default_max_dim = 15
        default_max_vines = 15
        min_granularity = 4
        max_granularity = 512
        min_granularity_vineyard = 1
        max_granularity_vineyard = max_granularity
        defr = 6
        default_x_start_first_line = (default_min_s + default_max_s) * (1 / defr)
        default_y_start_first_line = (default_min_k + default_max_k) * (1 / 2)
        default_x_end_first_line = (default_max_s + default_min_s) * (1 / 2)
        default_y_end_first_line = (default_min_k + default_max_k) * (1 / defr)
        default_x_start_second_line = (default_min_s + default_max_s) * (1 / 2)
        default_y_start_second_line = (default_min_k + default_max_k) * (
            (defr - 1) / defr
        )
        default_x_end_second_line = (default_max_s + default_min_s) * (
            (defr - 1) / defr
        )
        default_y_end_second_line = (default_min_k + default_max_k) * (1 / 2)

        self._app.title = "Persistable"
        self._app.layout = html.Div(
            className="root",
            children=[
                # contains the component counting function as a dictionary of lists
                dcc.Store(id=STORED_CCF),
                # contains the basic component counting function plot as a plotly figure
                dcc.Store(id=STORED_CCF_DRAWING),
                # contains the vineyard as a vineyard object
                dcc.Store(id=STORED_PV),
                # contains the basic prominence vineyard plot as a plotly figure
                dcc.Store(id=STORED_PV_DRAWING),
                #
                dcc.Store(id=STORED_PD, data=json.dumps([])),
                #
                dcc.Store(id=STORED_CCF_COMPUTATION_WARNINGS, data=json.dumps(" ")),
                dcc.Store(id=STORED_PV_COMPUTATION_WARNINGS, data=json.dumps(" ")),
                # dcc.Store(id=STORED_PD_COMPUTATION_WARNINGS),
                #
                dcc.Store(id=FIXED_PARAMETERS, data=json.dumps([])),
                #
                html.Div(id=EXPORTED_PARAMETER, hidden=True),
                #html.Div(id=EXPORTED_PARAMETER_2, hidden=True),
                html.Div(
                    className="grid",
                    children=[
                        html.Div(
                            children=[
                                html.H2("Component Counting Function"),
                                html.Div(
                                    className="parameters",
                                    children=[
                                        html.Details(
                                            id = CCF_DETAILS,
                                            children = [
                                                html.Summary("Inputs"),
                                                html.Div(
                                                    className="parameters",
                                                    children=[
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Distance scale min/max",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=MIN_DIST_SCALE,
                                                                    type="number",
                                                                    value=default_min_s,
                                                                    min=0,
                                                                    #step=default_s_step,
                                                                    debounce=True,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=MAX_DIST_SCALE,
                                                                    type="number",
                                                                    value=default_max_s,
                                                                    min=0,
                                                                    #step=default_s_step,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Density threshold min/max",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=MIN_DENSITY_THRESHOLD,
                                                                    type="number",
                                                                    value=default_min_k,
                                                                    min=0,
                                                                    #step=default_k_step,
                                                                    debounce=True,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=MAX_DENSITY_THRESHOLD,
                                                                    type="number",
                                                                    value=default_max_k,
                                                                    min=0,
                                                                    #step=default_k_step,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-single",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Granularity",
                                                                ),
                                                                dcc.Input(
                                                                    id=INPUT_GRANULARITY_CCF,
                                                                    className="small-value",
                                                                    type="number",
                                                                    value=default_granularity_ccf,
                                                                    min=min_granularity,
                                                                    max=max_granularity,
                                                                    debounce=True,
                                                                ),
                                                                html.Span(
                                                                    className="name",
                                                                    children="Max connected components",
                                                                ),
                                                                dcc.Input(
                                                                    id=INPUT_MAX_COMPONENTS,
                                                                    type="number",
                                                                    value=default_max_dim,
                                                                    min=1,
                                                                    className="small-value",
                                                                    step=1,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-single",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Number of cores computation",
                                                                ),
                                                                dcc.Input(
                                                                    className="small-value",
                                                                    id=INPUT_NUM_JOBS_CCF,
                                                                    type="number",
                                                                    value=default_num_jobs,
                                                                    min=1,
                                                                    step=1,
                                                                    max=16,
                                                                    debounce=True,
                                                                ),
                                                                html.Div(
                                                                    className="space"
                                                                ),
                                                            ],
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        html.Div(
                                            className="large-buttons",
                                            children=[
                                                html.Button(
                                                    "Compute",
                                                    id=COMPUTE_CCF_BUTTON,
                                                    className="button1",
                                                ),
                                                html.Button(
                                                    "Stop computation",
                                                    id=STOP_COMPUTE_CCF_BUTTON,
                                                    className="button2",
                                                    disabled=True,
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        html.Div(
                            children=[
                                html.H2("Prominence Vineyard"),
                                html.Div(
                                    className="parameters",
                                    children=[
                                        html.Details(
                                            id = PV_DETAILS,
                                            children = [
                                                html.Summary("Inputs"),
                                                html.Div(
                                                    className="parameters",
                                                    children=[
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="1st line start x/y",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=X_START_FIRST_LINE,
                                                                    type="number",
                                                                    value=default_x_start_first_line,
                                                                    min=0,
                                                                    # step=default_s_step,
                                                                    debounce=True,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=Y_START_FIRST_LINE,
                                                                    type="number",
                                                                    value=default_y_start_first_line,
                                                                    min=0,
                                                                    # step=default_k_step,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="1st line end x/y",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=X_END_FIRST_LINE,
                                                                    type="number",
                                                                    value=default_x_end_first_line,
                                                                    min=0,
                                                                    # step=default_s_step,
                                                                    debounce=True,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=Y_END_FIRST_LINE,
                                                                    type="number",
                                                                    value=default_y_end_first_line,
                                                                    min=0,
                                                                    # step=default_k_step,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="2nd line start x/y",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=X_START_SECOND_LINE,
                                                                    type="number",
                                                                    value=default_x_start_second_line,
                                                                    min=0,
                                                                    # step=default_s_step,
                                                                    debounce=True,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=Y_START_SECOND_LINE,
                                                                    type="number",
                                                                    value=default_y_start_second_line,
                                                                    min=0,
                                                                    # step=default_k_step,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-double",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="2nd line end x/y",
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=X_END_SECOND_LINE,
                                                                    type="number",
                                                                    value=default_x_end_second_line,
                                                                    min=0,
                                                                    # step=default_s_step,
                                                                ),
                                                                dcc.Input(
                                                                    className=VALUE,
                                                                    id=Y_END_SECOND_LINE,
                                                                    type="number",
                                                                    value=default_y_end_second_line,
                                                                    min=0,
                                                                    # step=default_k_step,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-single",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Number of lines vineyard",
                                                                ),
                                                                dcc.Input(
                                                                    id=INPUT_GRANULARITY_PV,
                                                                    className="small-value",
                                                                    type="number",
                                                                    value=default_granularity_pv,
                                                                    min=min_granularity_vineyard,
                                                                    max=max_granularity_vineyard,
                                                                    debounce=True,
                                                                ),
                                                                html.Span(
                                                                    className="name",
                                                                    children="Max number vines to display",
                                                                ),
                                                                dcc.Input(
                                                                    id=INPUT_MAX_VINES,
                                                                    type="number",
                                                                    value=default_max_vines,
                                                                    min=1,
                                                                    className="small-value",
                                                                    step=1,
                                                                    debounce=True,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Div(
                                                            className="parameter-single",
                                                            children=[
                                                                html.Span(
                                                                    className="name",
                                                                    children="Number of cores computation",
                                                                ),
                                                                dcc.Input(
                                                                    className="small-value",
                                                                    id=INPUT_NUM_JOBS_PV,
                                                                    type="number",
                                                                    value=default_num_jobs,
                                                                    min=1,
                                                                    step=1,
                                                                    max=16,
                                                                    debounce=True,
                                                                ),
                                                                html.Span(
                                                                    className="name",
                                                                    children="Prominence scale",
                                                                ),
                                                                dcc.RadioItems(
                                                                    ["Lin", "Log"],
                                                                    "Lin",
                                                                    id=INPUT_PROM_VIN_SCALE,
                                                                    className="small-value",
                                                                ),
                                                            ],
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        html.Div(
                                            className="parameters",
                                            children=[
                                                html.Div(
                                                    className="large-buttons",
                                                    children=[
                                                        html.Button(
                                                            "Compute",
                                                            id=COMPUTE_PV_BUTTON,
                                                            className="button1",
                                                        ),
                                                        html.Button(
                                                            "Stop computation",
                                                            id=STOP_COMPUTE_PV_BUTTON,
                                                            className="button2",
                                                            disabled=True,
                                                        ),
                                                    ],
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        dcc.Graph(
                            id=CCF_PLOT,
                            className="graph",
                            figure=empty_figure(),
                            config={
                                "responsive": True,
                                "displayModeBar": False,
                                "modeBarButtonsToRemove": [
                                    "toImage",
                                    "pan",
                                    "zoomIn",
                                    "zoomOut",
                                    "resetScale",
                                    "lasso",
                                    "select",
                                ],
                                "displaylogo": False,
                            },
                        ),
                        dcc.Graph(
                            id=PV_PLOT,
                            className="graph",
                            figure=empty_figure(),
                            config={
                                "responsive": True,
                                "displayModeBar": False,
                                "modeBarButtonsToRemove": [
                                    "toImage",
                                    "pan",
                                    "zoomIn",
                                    "zoomOut",
                                    "resetScale",
                                    "lasso",
                                    "select",
                                ],
                                "displaylogo": False,
                            },
                        ),
                        html.Div(
                            className="plot-tools",
                            children=[
                                html.Div(
                                    id=CCF_PLOT_CONTROLS_DIV,
                                    className="parameters",
                                    hidden=True,
                                    children=[
                                        html.Div(
                                            className="parameter-single",
                                            children=[
                                                html.Span(
                                                    className="name",
                                                    children="Vineyard inputs selection",
                                                ),
                                                dcc.RadioItems(
                                                    ["On", "Off"],
                                                    "Off",
                                                    id=DISPLAY_LINES_SELECTION,
                                                    className=VALUE,
                                                ),
                                            ],
                                        ),
                                        html.Div(
                                            className="parameter-single",
                                            id=ENDPOINT_SELECTION_DIV,
                                            children=[
                                                html.Span(
                                                    className="name",
                                                    children="Endpoint",
                                                ),
                                                dcc.RadioItems(
                                                    [
                                                        "1st line start",
                                                        "1st line end",
                                                        "2nd line start",
                                                        "2nd line end",
                                                    ],
                                                    "1st line start",
                                                    id=ENDPOINT_SELECTION,
                                                    className=VALUE,
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        html.Div(
                            className="plot-tools",
                            children=[
                                html.Div(
                                    id=PV_PLOT_CONTROLS_DIV,
                                    className="parameters",
                                    hidden=True,
                                    children=[
                                        html.Div(
                                            className="parameter-single",
                                            children=[
                                                html.Span(
                                                    className="name",
                                                    children="Parameter selection",
                                                ),
                                                dcc.RadioItems(
                                                    ["On", "Off"],
                                                    "Off",
                                                    id=DISPLAY_PARAMETER_SELECTION,
                                                    className=VALUE,
                                                ),
                                            ],
                                        ),
                                        html.Div(
                                            className="parameter-double-button",
                                            id=PARAMETER_SELECTION_DIV,
                                            children=[
                                                html.Span(
                                                    className="name",
                                                    children="Line number",
                                                ),
                                                dcc.Input(
                                                    className=VALUE,
                                                    id=INPUT_LINE,
                                                    type="number",
                                                    value=1,
                                                    min=1,
                                                    debounce=True,
                                                ),
                                                html.Span(
                                                    className="name",
                                                    children="Gap number",
                                                ),
                                                dcc.Input(
                                                    className=VALUE,
                                                    id=INPUT_GAP,
                                                    type="number",
                                                    value=1,
                                                    min=1,
                                                    debounce=True,
                                                ),
                                                html.Button(
                                                    "Choose parameter",
                                                    id=EXPORT_PARAMETERS_BUTTON,
                                                    className="button",
                                                    disabled=True,
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
                html.Details(
                    id=LOG_DIV,
                    className="warnings",
                    children=[
                        html.Summary("Warnings"),
                        html.Pre(
                            id=LOG,
                            children=[" "],
                        ),
                    ],
                    open=False,
                ),
                html.Details(
                    className="help",
                    children=[
                        html.Summary("Tips"),
                        dcc.Markdown(
                            """
            - Check the documentation at [persistable.readthedocs.io](https://persistable.readthedocs.io/).
            - Check the log, above, for warnings.
            - Press enter after you modify numerical fields.
            - Computing the component counting function and prominence vineyard can take a while, depending on the size and dimensionality of the dataset as well as other factors.
            - Make sure to leave your pointer still when clicking on the component counting function plot, otherwise your interaction may not be registered.
            """
                        ),
                    ],
                ),
            ],
        )


    # when using long callbacks, dash will pickle all the objects that are used
    # in the function that is being registered as a long callback. In particular,
    # using self._persistable will also pickle self, which in this case contains
    # also a dash app, which cannot be pickled. We get around this by passing
    # self._persistable to the _register_callbacks function.
    def _register_callbacks(self, persistable, debug):
        def dash_callback(
            inputs,
            outputs,
            prevent_initial_call=False,
            background=False,
            running=None,
            cancel=None,
        ):
            def cs(l):
                return l[0] + l[1]

            def out_function(function):
                dash_inputs = [
                    dash.Input(i, v) if t == IN else dash.State(i, v)
                    for i, v, t in inputs
                ]
                dash_outputs = (
                    [dash.Output(i, v) for i, v in outputs]
                    if len(outputs) > 1
                    else dash.Output(outputs[0][0], outputs[0][1])
                )

                if running is None:
                    dash_running_outputs = None
                else:
                    dash_running_outputs = [
                        (dash.Output(i, v), value_start, value_end)
                        for i, v, value_start, value_end in running
                    ]

                if cancel is None:
                    dash_cancel = None
                else:
                    dash_cancel = [(dash.Input(i, v)) for i, v in cancel]

                def callback_function(*argv):
                    d = {}
                    for n, arg in enumerate(argv):
                        if arg is None:
                            raise PreventUpdate
                        d[cs(inputs[n])] = arg
                    d = function(d)
                    return (
                        tuple(d[cs(o)] for o in outputs)
                        if len(outputs) > 1
                        else d[cs(outputs[0])]
                    )

                if background:
                    self._app.long_callback(
                        dash_outputs,
                        dash_inputs,
                        prevent_initial_call,
                        running=dash_running_outputs,
                        cancel=dash_cancel,
                    )(callback_function)
                    # TODO: figure out why the following causes problems with joblib Parallel
                    # Note that callback with background=True is the recommended way of running
                    # background jobs in dash now.
                    #self._app.callback(
                    #    dash_outputs,
                    #    dash_inputs,
                    #    prevent_initial_call,
                    #    running=dash_running_outputs,
                    #    cancel=dash_cancel,
                    #    background=True,
                    #)(callback_function)
                else:
                    self._app.callback(
                        dash_outputs,
                        dash_inputs,
                        prevent_initial_call,
                    )(callback_function)

                return function

            return out_function


        @dash_callback(
            [[DISPLAY_PARAMETER_SELECTION, VALUE, IN]],
            [[PARAMETER_SELECTION_DIV, HIDDEN]],
        )
        def toggle_parameter_selection_pv(d):
            if d[DISPLAY_PARAMETER_SELECTION + VALUE] == "On":
                d[PARAMETER_SELECTION_DIV + HIDDEN] = False
            else:
                d[PARAMETER_SELECTION_DIV + HIDDEN] = True
            return d

        @dash_callback(
            [
                [STORED_CCF_COMPUTATION_WARNINGS, DATA, IN],
                [STORED_PV_COMPUTATION_WARNINGS, DATA, IN],
            ],
            [[LOG, CHILDREN], [LOG_DIV, "open"]],
            True,
        )
        def print_log(d):
            if ctx.triggered_id == STORED_CCF_COMPUTATION_WARNINGS:
                message = json.loads(d[STORED_CCF_COMPUTATION_WARNINGS + DATA])
            elif ctx.triggered_id == STORED_PV_COMPUTATION_WARNINGS:
                message = json.loads(d[STORED_PV_COMPUTATION_WARNINGS + DATA])
            else:
                raise Exception(
                    "print_log was triggered by unknown id: " + str(ctx.triggered_id)
                )
            if len(message) > 0:
                d[LOG_DIV + "open"] = True
                d[LOG + CHILDREN] = message
            else:
                d[LOG_DIV + "open"] = False
                d[LOG + CHILDREN] = " "

            return d

        @dash_callback(
            [
                [CCF_PLOT, CLICKDATA, IN],
                [DISPLAY_LINES_SELECTION, VALUE, ST],
                [ENDPOINT_SELECTION, VALUE, ST],
                [X_START_FIRST_LINE, VALUE, ST],
                [Y_START_FIRST_LINE, VALUE, ST],
                [X_END_FIRST_LINE, VALUE, ST],
                [Y_END_FIRST_LINE, VALUE, ST],
                [X_START_SECOND_LINE, VALUE, ST],
                [Y_START_SECOND_LINE, VALUE, ST],
                [X_END_SECOND_LINE, VALUE, ST],
                [Y_END_SECOND_LINE, VALUE, ST],
            ],
            [
                [X_START_FIRST_LINE, VALUE],
                [Y_START_FIRST_LINE, VALUE],
                [X_END_FIRST_LINE, VALUE],
                [Y_END_FIRST_LINE, VALUE],
                [X_START_SECOND_LINE, VALUE],
                [Y_START_SECOND_LINE, VALUE],
                [X_END_SECOND_LINE, VALUE],
                [Y_END_SECOND_LINE, VALUE],
            ],
            True,
        )
        def on_ccf_click(d):
            if d[DISPLAY_LINES_SELECTION + VALUE] == "On":
                new_x, new_y = (
                    d[CCF_PLOT + CLICKDATA]["points"][0]["x"],
                    d[CCF_PLOT + CLICKDATA]["points"][0]["y"],
                )
                if d[ENDPOINT_SELECTION + VALUE] == "1st line start":
                    d[X_START_FIRST_LINE + VALUE] = new_x
                    d[Y_START_FIRST_LINE + VALUE] = new_y
                elif d[ENDPOINT_SELECTION + VALUE] == "1st line end":
                    d[X_END_FIRST_LINE + VALUE] = new_x
                    d[Y_END_FIRST_LINE + VALUE] = new_y
                elif d[ENDPOINT_SELECTION + VALUE] == "2nd line start":
                    d[X_START_SECOND_LINE + VALUE] = new_x
                    d[Y_START_SECOND_LINE + VALUE] = new_y
                elif d[ENDPOINT_SELECTION + VALUE] == "2nd line end":
                    d[X_END_SECOND_LINE + VALUE] = new_x
                    d[Y_END_SECOND_LINE + VALUE] = new_y
            return d

        @dash_callback(
            [[DISPLAY_LINES_SELECTION, VALUE, IN]], [[ENDPOINT_SELECTION_DIV, HIDDEN]]
        )
        def toggle_parameter_selection_ccf(d):
            if d[DISPLAY_LINES_SELECTION + VALUE] == "On":
                d[ENDPOINT_SELECTION_DIV + HIDDEN] = False
            else:
                d[ENDPOINT_SELECTION_DIV + HIDDEN] = True
            return d

        @dash_callback(
            [[STORED_CCF, DATA, IN], [INPUT_MAX_COMPONENTS, VALUE, IN]],
            [[STORED_CCF_DRAWING, DATA]],
            False,
        )
        def draw_ccf(d):
            ccf = d[STORED_CCF + DATA]

            ccf = json.loads(ccf)

            max_components = d[INPUT_MAX_COMPONENTS + VALUE]

            fig = go.Figure(
                layout=go.Layout(
                    xaxis_title="Distance scale",
                    yaxis_title="Density threshold",
                    xaxis={"fixedrange": True},
                    yaxis={"fixedrange": True},
                ),
            )

            fig.add_trace(
                go.Heatmap(
                    **ccf,
                    hovertemplate="<b># comp.: %{z:d}</b><br>x: %{x:.3e} <br>y: %{y:.3e} ",
                    zmin=0,
                    zmax=max_components,
                    showscale=False,
                    name="",
                )
            )
            fig.update_traces(colorscale="greys")
            fig.update_layout(showlegend=False)
            fig.update_layout(autosize=True)
            fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))
            fig.update_layout(clickmode="event+select")

            d[STORED_CCF_DRAWING + DATA] = plotly.io.to_json(fig)
            return d

        @dash_callback(
            [
                [STORED_CCF_DRAWING, DATA, IN],
                [MIN_DIST_SCALE, VALUE, IN],
                [MAX_DIST_SCALE, VALUE, IN],
                [MIN_DENSITY_THRESHOLD, VALUE, IN],
                [MAX_DENSITY_THRESHOLD, VALUE, IN],
                [DISPLAY_LINES_SELECTION, VALUE, IN],
                [X_START_FIRST_LINE, VALUE, IN],
                [Y_START_FIRST_LINE, VALUE, IN],
                [X_END_FIRST_LINE, VALUE, IN],
                [Y_END_FIRST_LINE, VALUE, IN],
                [X_START_SECOND_LINE, VALUE, IN],
                [Y_START_SECOND_LINE, VALUE, IN],
                [X_END_SECOND_LINE, VALUE, IN],
                [Y_END_SECOND_LINE, VALUE, IN],
                [ENDPOINT_SELECTION, VALUE, IN],
                [DISPLAY_PARAMETER_SELECTION, VALUE, IN],
                [FIXED_PARAMETERS, DATA, IN],
                [STORED_PD, DATA, ST],
                [INPUT_GAP, VALUE, ST],
            ],
            [[CCF_PLOT, FIGURE]],
            False,
        )
        def draw_ccf_extras(d):
            fig = plotly.io.from_json(d[STORED_CCF_DRAWING + DATA])

            def generate_line(
                xs, ys, text, color="mediumslateblue", different_marker=None
            ):
                if different_marker == None:
                    marker_styles = ["circle", "circle"]
                elif different_marker == 0:
                    marker_styles = ["circle-open", "circle"]
                elif different_marker == 1:
                    marker_styles = ["circle", "circle-open"]
                return go.Scatter(
                    x=xs,
                    y=ys,
                    name=text + " line",
                    text=[text + " line start", text + " line end"],
                    marker=dict(size=15, color=color),
                    marker_symbol=marker_styles,
                    hoverinfo="skip",
                    showlegend=False,
                    mode="markers+lines+text",
                    textposition=["top center", "bottom center"],
                )

            if d[DISPLAY_LINES_SELECTION + VALUE] == "On":

                # draw polygon
                fig.add_trace(
                    go.Scatter(
                        x=[
                            d[X_START_FIRST_LINE + VALUE],
                            d[X_END_FIRST_LINE + VALUE],
                            d[X_END_SECOND_LINE + VALUE],
                            d[X_START_SECOND_LINE + VALUE],
                        ],
                        y=[
                            d[Y_START_FIRST_LINE + VALUE],
                            d[Y_END_FIRST_LINE + VALUE],
                            d[Y_END_SECOND_LINE + VALUE],
                            d[Y_START_SECOND_LINE + VALUE],
                        ],
                        fillcolor="rgba(0, 0, 255, 0.05)",
                        fill="toself",
                        mode="none",
                        hoverinfo="skip",
                        showlegend=False,
                        # text=text,
                        name="",
                    )
                )

                if d[ENDPOINT_SELECTION + VALUE] == "1st line start":
                    first_line_endpoints = 0
                    second_line_endpoints = None
                elif d[ENDPOINT_SELECTION + VALUE] == "1st line end":
                    first_line_endpoints = 1
                    second_line_endpoints = None
                elif d[ENDPOINT_SELECTION + VALUE] == "2nd line start":
                    first_line_endpoints = None
                    second_line_endpoints = 0
                elif d[ENDPOINT_SELECTION + VALUE] == "2nd line end":
                    first_line_endpoints = None
                    second_line_endpoints = 1

                fig.add_trace(
                    generate_line(
                        [d[X_START_FIRST_LINE + VALUE], d[X_END_FIRST_LINE + VALUE]],
                        [d[Y_START_FIRST_LINE + VALUE], d[Y_END_FIRST_LINE + VALUE]],
                        "1st",
                        different_marker=first_line_endpoints,
                        color="mediumslateblue",
                    )
                )
                fig.add_trace(
                    generate_line(
                        [d[X_START_SECOND_LINE + VALUE], d[X_END_SECOND_LINE + VALUE]],
                        [d[Y_START_SECOND_LINE + VALUE], d[Y_END_SECOND_LINE + VALUE]],
                        "2nd",
                        different_marker=second_line_endpoints,
                        color="mediumslateblue",
                    )
                )

            params = json.loads(d[FIXED_PARAMETERS + DATA])
            if len(params) != 0 and d[DISPLAY_PARAMETER_SELECTION + VALUE] == "On":
                pd = json.loads(d[STORED_PD + DATA])
                if len(pd) != 0:

                    def generate_bar(xs, ys, color):
                        return go.Scatter(
                            x=xs,
                            y=ys,
                            marker=dict(color=color),
                            hoverinfo="skip",
                            showlegend=False,
                            mode="lines",
                            line=dict(width=6),
                        )

                    st_x = params["start"][0]
                    st_y = params["start"][1]
                    end_x = params["end"][0]
                    end_y = params["end"][1]
                    st = np.array([st_x, st_y])
                    end = np.array([end_x, end_y])
                    A = end - st

                    # ideally we would get the actual ratio of the rendered picture
                    # we are using an estimate given by the "usual" way in which persistable
                    # is rendered
                    ratio = np.array([3,2])
                    ratio = (ratio / np.linalg.norm(ratio)) * np.linalg.norm(np.array([1,1]))
                    alpha = [
                        (d[MAX_DIST_SCALE + VALUE] - d[MIN_DIST_SCALE + VALUE]),
                        ( d[MAX_DENSITY_THRESHOLD + VALUE] - d[MIN_DENSITY_THRESHOLD + VALUE]),
                    ]
                    alpha = np.array(alpha) / ratio

                    A = A/alpha
                    B = np.array([-A[1], A[0]])

                    shift = 50
                    B = B / np.linalg.norm(B)
                    tau = B * alpha/ shift

                    pd = np.array(pd)
                    lengths = pd[:, 1] - pd[:, 0]
                    pd = pd[np.argsort(lengths)[::-1]]
                    for i, point in enumerate(pd):
                        l = end_x - st_x
                        p_st = point[0]
                        p_end = point[1]
                        q_st_x = st_x + p_st
                        q_end_x = st_x + p_end
                        q_st_y = st_y - (st_y - end_y) * (p_st / l)
                        q_end_y = st_y - (st_y - end_y) * (p_end / l)
                        q_st = np.array([q_st_x, q_st_y])
                        q_end = np.array([q_end_x, q_end_y])
                        r_st = q_st + (i + 1) * tau
                        r_end = q_end + (i + 1) * tau
                        color = (
                            "rgba(34, 139, 34, 1)"
                            if i < d[INPUT_GAP + VALUE]
                            else "rgba(34, 139, 34, 0.3)"
                        )
                        fig.add_trace(
                            generate_bar(
                                [r_st[0], r_end[0]], [r_st[1], r_end[1]], color
                            )
                        )
                fig.add_trace(
                    generate_line(
                        [st_x, end_x],
                        [st_y, end_y],
                        "selected",
                        color="blue",
                    )
                )

            fig.update_layout(
                xaxis=dict(range=[d[MIN_DIST_SCALE + VALUE], d[MAX_DIST_SCALE + VALUE]]),
                yaxis=dict(range=[
                    d[MIN_DENSITY_THRESHOLD + VALUE],
                    d[MAX_DENSITY_THRESHOLD + VALUE],
                ])
            )

            d[CCF_PLOT + FIGURE] = fig
            return d

        @dash_callback(
            [[STORED_PV, DATA, ST], [STORED_PV_DRAWING, DATA, IN]], [[PV_PLOT, FIGURE]]
        )
        def draw_pv_post(d):

            d[PV_PLOT + FIGURE] = plotly.io.from_json(d[STORED_PV_DRAWING + DATA])

            return d

        @dash_callback(
            [
                [COMPUTE_CCF_BUTTON, N_CLICKS, IN],
                [MIN_DENSITY_THRESHOLD, VALUE, ST],
                [MAX_DENSITY_THRESHOLD, VALUE, ST],
                [MIN_DIST_SCALE, VALUE, ST],
                [MAX_DIST_SCALE, VALUE, ST],
                [INPUT_GRANULARITY_CCF, VALUE, ST],
                [INPUT_NUM_JOBS_CCF, VALUE, ST],
            ],
            [
                [STORED_CCF, DATA],
                [STORED_CCF_COMPUTATION_WARNINGS, DATA],
                [CCF_PLOT_CONTROLS_DIV, HIDDEN],
            ],
            prevent_initial_call=True,
            background=True,
            running=[
                [COMPUTE_CCF_BUTTON, DISABLED, True, False],
                [STOP_COMPUTE_CCF_BUTTON, DISABLED, False, True],
            ],
            cancel=[[STOP_COMPUTE_CCF_BUTTON, N_CLICKS]],
        )
        def compute_ccf(d):
            if debug:
                print("Compute ccf in background started.")

            granularity = d[INPUT_GRANULARITY_CCF + VALUE]
            num_jobs = int(d[INPUT_NUM_JOBS_CCF + VALUE])

            out = ""
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                try:
                    ss, ks, hf = persistable._compute_hilbert_function(
                        d[MIN_DENSITY_THRESHOLD + VALUE],
                        d[MAX_DENSITY_THRESHOLD + VALUE],
                        d[MIN_DIST_SCALE + VALUE],
                        d[MAX_DIST_SCALE + VALUE],
                        granularity,
                        n_jobs=num_jobs,
                    )
                except ValueError:
                    out += traceback.format_exc()
                    d[STORED_CCF_COMPUTATION_WARNINGS + DATA] = json.dumps(out)
                    d[STORED_CCF + DATA] = None
                    d[CCF_PLOT_CONTROLS_DIV + HIDDEN] = True
                    return d

            for a in w:
                out += warnings.formatwarning(
                    a.message, a.category, a.filename, a.lineno
                )

            d[STORED_CCF + DATA] = json.dumps(
                {"y": ks[:-1].tolist(), "x": ss[:-1].tolist(), "z" : hf.tolist()}
            )

            d[STORED_CCF_COMPUTATION_WARNINGS + DATA] = json.dumps(out)
            d[CCF_PLOT_CONTROLS_DIV + HIDDEN] = False

            if debug:
                print("Compute ccf in background finished.")

            return d

        @dash_callback(
            [
                [COMPUTE_PV_BUTTON, N_CLICKS, IN],
                [X_START_FIRST_LINE, VALUE, ST],
                [Y_START_FIRST_LINE, VALUE, ST],
                [X_END_FIRST_LINE, VALUE, ST],
                [Y_END_FIRST_LINE, VALUE, ST],
                [X_START_SECOND_LINE, VALUE, ST],
                [Y_START_SECOND_LINE, VALUE, ST],
                [X_END_SECOND_LINE, VALUE, ST],
                [Y_END_SECOND_LINE, VALUE, ST],
                [INPUT_GRANULARITY_PV, VALUE, ST],
                [INPUT_NUM_JOBS_PV, VALUE, ST],
            ],
            [
                [STORED_PV, DATA],
                [STORED_PV_COMPUTATION_WARNINGS, DATA],
                [INPUT_LINE, "max"],
                [INPUT_LINE, VALUE],
                [EXPORT_PARAMETERS_BUTTON, DISABLED],
                [PV_PLOT_CONTROLS_DIV, HIDDEN],
            ],
            prevent_initial_call=True,
            background=True,
            running=[
                [COMPUTE_PV_BUTTON, DISABLED, True, False],
                [STOP_COMPUTE_PV_BUTTON, DISABLED, False, True],
            ],
            cancel=[[STOP_COMPUTE_PV_BUTTON, N_CLICKS]],
        )
        def compute_pv(d):
            if debug:
                print("Compute pv in background started.")

            granularity = d[INPUT_GRANULARITY_PV + VALUE]
            num_jobs = int(d[INPUT_NUM_JOBS_PV + VALUE])

            out = ""
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                try:
                    pv = persistable._compute_vineyard(
                        [
                            [
                                d[X_START_FIRST_LINE + VALUE],
                                d[Y_START_FIRST_LINE + VALUE],
                            ],
                            [d[X_END_FIRST_LINE + VALUE], d[Y_END_FIRST_LINE + VALUE]],
                        ],
                        [
                            [
                                d[X_START_SECOND_LINE + VALUE],
                                d[Y_START_SECOND_LINE + VALUE],
                            ],
                            [
                                d[X_END_SECOND_LINE + VALUE],
                                d[Y_END_SECOND_LINE + VALUE],
                            ],
                        ],
                        n_parameters=granularity,
                        n_jobs=num_jobs,
                    )
                except ValueError:
                    out += traceback.format_exc()
                    d[STORED_PV_COMPUTATION_WARNINGS + DATA] = json.dumps(out)
                    d[STORED_PV + DATA] = None
                    d[INPUT_LINE + "max"] = granularity
                    d[INPUT_LINE + VALUE] = granularity // 2
                    d[EXPORT_PARAMETERS_BUTTON + DISABLED] = True
                    d[PV_PLOT_CONTROLS_DIV + HIDDEN] = True
                    return d

            for a in w:
                out += warnings.formatwarning(
                    a.message, a.category, a.filename, a.lineno
                )

            d[STORED_PV + DATA] = json.dumps(pv.__dict__)
            d[STORED_PV_COMPUTATION_WARNINGS + DATA] = json.dumps(out)
            d[INPUT_LINE + "max"] = granularity
            d[INPUT_LINE + VALUE] = granularity // 2
            d[EXPORT_PARAMETERS_BUTTON + DISABLED] = False

            d[PV_PLOT_CONTROLS_DIV + HIDDEN] = False

            if debug:
                print("Compute pv in background finished.")

            return d

        @dash_callback(
            [
                [STORED_PV, DATA, IN],
                [INPUT_MAX_VINES, VALUE, IN],
                [INPUT_PROM_VIN_SCALE, VALUE, IN],
                [DISPLAY_PARAMETER_SELECTION, VALUE, IN],
                [INPUT_LINE, VALUE, IN],
                [INPUT_GAP, VALUE, IN],
            ],
            [[STORED_PV_DRAWING, DATA]],
            False,
        )
        def draw_pv(d):
            firstn = d[INPUT_MAX_VINES + VALUE]

            vineyard_as_dict = json.loads(d[STORED_PV + DATA])
            vineyard = Vineyard(
                vineyard_as_dict["_parameters"],
                vineyard_as_dict["_persistence_diagrams"],
            )

            _vineyard_values = []

            times = np.array(vineyard.parameter_indices()) + 1
            vines = vineyard._vineyard_to_vines()
            num_vines = min(len(vines), firstn)

            fig = go.Figure(
                layout=go.Layout(
                    xaxis_title="Parameter line",
                    yaxis_title="Prominence",
                    xaxis={
                        "fixedrange": True,
                        "showgrid": False,
                    },
                    yaxis={
                        "fixedrange": True,
                        "showgrid": False,
                    },
                    plot_bgcolor="rgba(0, 0, 0, 0.1)",
                ),
            )

            colors = sample_colorscale(
                "viridis", list(np.linspace(0, 1, num_vines))[::-1]
            )
            if num_vines > 0:
                for i in range(num_vines - 1, -1, -1):
                    till = "tozeroy" if i == num_vines - 1 else "tonexty"
                    color = colors[i]
                    if (
                        d[DISPLAY_PARAMETER_SELECTION + VALUE] == "On"
                        and i + 1 == d[INPUT_GAP + VALUE]
                    ):
                        fig.add_trace(
                            go.Scatter(
                                x=times,
                                y=vines[i][1],
                                fill=till,
                                # hoveron="fills",
                                text="vine " + str(i + 1),
                                hoverinfo="text",
                                line_color=color,
                                fillpattern=go.scatter.Fillpattern(shape="x"),
                            )
                        )
                    else:
                        fig.add_trace(
                            go.Scatter(
                                x=times,
                                y=vines[i][1],
                                fill=till,
                                # hoveron="fills",
                                text="vine " + str(i + 1),
                                hoverinfo="text",
                                line_color=color,
                            )
                        )
            for i, tv in enumerate(vines):
                times, vine = tv
                for vine_part, time_part in vineyard._vine_parts(vine):
                    _vineyard_values.extend(vine_part)
            values = np.array(_vineyard_values)

            if d[DISPLAY_PARAMETER_SELECTION + VALUE] == "On":
                fig.add_vline(x=d[INPUT_LINE + VALUE], line_color="grey")

            if len(values) > 0:
                if d[INPUT_PROM_VIN_SCALE + VALUE] == "Log":
                    fig.update_layout(yaxis_type="log")
                    fig.update_layout(
                        yaxis_range=[
                            np.log10(np.quantile(values[values > 0], 0.05)),
                            np.log10(max(values)),
                        ]
                    )
                else:
                    fig.update_layout(yaxis_range=[0, max(values)])

            fig.update_layout(showlegend=False)
            fig.update_layout(autosize=True)
            fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))

            d[STORED_PV_DRAWING + DATA] = plotly.io.to_json(fig)

            return d

        @dash_callback(
            [
                [INPUT_GAP, VALUE, IN],
                [INPUT_LINE, VALUE, IN],
                [STORED_PV, DATA, IN],
            ],
            [
                [FIXED_PARAMETERS, DATA],
                [STORED_PD, DATA],
            ],
            True,
        )
        def fix_parameters(d):
            vineyard_as_dict = json.loads(d[STORED_PV + DATA])
            vineyard = Vineyard(
                vineyard_as_dict["_parameters"],
                vineyard_as_dict["_persistence_diagrams"],
            )
            line = vineyard._parameters[d[INPUT_LINE + VALUE] - 1]
            params = {
                "n_clusters": d[INPUT_GAP + VALUE],
                "start": line[0],
                "end": line[1],
            }
            d[FIXED_PARAMETERS + DATA] = json.dumps(params)

            pd = vineyard._persistence_diagrams[d[INPUT_LINE + VALUE] - 1]

            d[STORED_PD + DATA] = json.dumps(pd)

            return d

        @dash_callback(
            [
                [EXPORT_PARAMETERS_BUTTON, N_CLICKS, IN],
                [FIXED_PARAMETERS, DATA, ST],
            ],
            [[EXPORTED_PARAMETER, CHILDREN]],
            True,
        )
        def export_parameters(d):
            self._parameters_sem.acquire()
            self._parameters = json.loads(d[FIXED_PARAMETERS + DATA])
            self._parameters_sem.release()
            d[EXPORTED_PARAMETER + CHILDREN] = json.dumps(self._parameters)
            return d



# Authors: Luis Scoccola
# License: 3-clause BSD

# from re import A
from turtle import back
import numpy as np
import warnings


import plotly.graph_objects as go
import plotly
from jupyter_dash import JupyterDash
import dash
from dash import dcc
from dash import html
from dash import DiskcacheManager, CeleryManager
from dash.long_callback import DiskcacheLongCallbackManager

import pandas as pd
import json
import diskcache

PERSISTABLE_STDERR = "./persistable-stderr"
PERSISTABLE_DASH_CACHE = "./persistable-dash-cache"
# WARNINGS_GLOBAL = "test default"

X_START_FIRST_LINE = "x-start-first-line-"
Y_START_FIRST_LINE = "y-start-first-line-"
X_END_FIRST_LINE = "x-end-first-line-"
Y_END_FIRST_LINE = "y-end-first-line-"
X_START_SECOND_LINE = "x-start-second-line-"
Y_START_SECOND_LINE = "y-start-second-line-"
X_END_SECOND_LINE = "x-end-second-line-"
Y_END_SECOND_LINE = "y-end-second-line-"
DISPLAY_LINES_SELECTION = "display-lines-selection-"
ENDPOINT_SELECTION = "endpoint-selection-"
MIN_DIST_SCALE = "min-dist-scale-"
MAX_DIST_SCALE = "max-dist-scale-"
MIN_DENSITY_THRESHOLD = "min-density-threshold-"
MAX_DENSITY_THRESHOLD = "max-density-threshold-"
ENDPOINT_SELECTION_DIV = "endpoint-selection-div-"
PARAMETER_SELECTION_DIV = "parameter-selection-div-"
DISPLAY_PARAMETER_SELECTION = "display-parameter-selection-"

CFF_PLOT = "cff-plot-"
STORED_CCF = "stored-ccf-"
STORED_CCF_DRAWING = "stored-ccf-drawing-"
COMPUTE_CCF_BUTTON = "compute-ccf-button-"
INPUT_LOG_GRANULARITY_CCF = "input-log-granularity-ccf-"
INPUT_NUM_JOBS_CCF = "input-num-jobs-ccf-"
INPUT_MAX_COMPONENTS = "input-max-components-"
LOG = "log-"
WARNINGS_POLLING_INTERVAL = "warnings-polling-interval-"

PV_PLOT = "pv-plot-"
STORED_PV = "stored-pv-"
STORED_PV_DRAWING = "stored-pv-drawing-"
COMPUTE_PV_BUTTON = "compute-pv-button-"
INPUT_MAX_VINES = "input-max-vines-"
INPUT_PROM_VIN_SCALE = "input-prom-vin-scale-"
INPUT_LOG_GRANULARITY_PV = "input-log-granularity-pv-"
INPUT_NUM_JOBS_PV = "input-num-jobs-pv-"

VALUE = "value"
CLICKDATA = "clickData"
HIDDEN = "hidden"
DATA = "data"
N_CLICKS = "n_clicks"
DISABLED = "disabled"
FIGURE = "figure"
CHILDREN = "children"
N_INTERVALS = "n_intervals"

IN = "input"
ST = "state"


def cs(l):
    return l[0] + l[1]


_funcs = []

COUNTER = 0

def my_callback(
    inputs, outputs, _app, prevent_initial_call=False, background=False, running=None, func_num = 0,
):

    def out_function(function):
        print("attaching callback to")
        print(function.__name__)
        # FIX
        dash_inputs = [
            dash.Input(i, v) if t == IN else dash.State(i, v) for i, v, t in inputs
        ]
        dash_outputs = (
            [dash.Output(i, v) for i, v in outputs]
            if len(outputs) > 1
            else dash.Output(outputs[0][0], outputs[0][1])
        )

        if running is None:
            dash_running_outputs = None
        else:
            dash_running_outputs = [
                (dash.Output(i, v), value_start, value_end)
                for i, v, value_start, value_end in running
            ]


        def callback_function(*argv):
            d = {}
            for n, arg in enumerate(argv):
                d[cs(inputs[n])] = arg
            d = function(d)
            return (
                tuple(d[cs(o)] for o in outputs)
                if len(outputs) > 1
                else d[cs(outputs[0])]
            )

        def callback_function_1(*argv):
            d = {}
            for n, arg in enumerate(argv):
                d[cs(inputs[n])] = arg
            d = function(d)
            return (
                tuple(d[cs(o)] for o in outputs)
                if len(outputs) > 1
                else d[cs(outputs[0])]
            )

        def callback_function_2(*argv):
            d = {}
            for n, arg in enumerate(argv):
                d[cs(inputs[n])] = arg
            d = function(d)
            return (
                tuple(d[cs(o)] for o in outputs)
                if len(outputs) > 1
                else d[cs(outputs[0])]
            )

        #global COUNTER
        #import inspect
        #import re
        #fn_source = inspect.getsource(callback_function) 
        #fn_str = fn_source 
        #print(fn_str.encode("utf-8"))
        #new = re.sub("callback_function", "callback_function_" + str(COUNTER), str(fn_str.encode("utf-8")))
        #callback_function = exec(new,globals(),locals())
        #print(str(inspect.getsource(callback_function).encode("utf-8"))) 

        ##locals()[func_label] = locals()["callback_function"]
        ##callback_function.__name__ = func_label
        #COUNTER += 1

        #fn_source = inspect.getsource(callback_function) 
        #fn_str = fn_source 
        #print(fn_str.encode("utf-8"))

        #fn_source = inspect.getsource(locals()[func_label]) 
        #fn_str = fn_source 
        #print(fn_str.encode("utf-8"))

        global _funcs
        _funcs.append(callback_function)
        _funcs.append(function)
        _funcs.append(inputs)
        _funcs.append(outputs)
        _funcs.append(dash_inputs)
        _funcs.append(dash_outputs)
        _funcs.append(dash_running_outputs)


        if background:
            if func_num==0:
                _app.callback(
                    dash_outputs,
                    dash_inputs,
                    prevent_initial_call,
                    running=dash_running_outputs,
                    background=background,
                )(callback_function_1)
            else:
                _app.callback(
                    dash_outputs,
                    dash_inputs,
                    prevent_initial_call,
                    running=dash_running_outputs,
                    background=background,
                )(callback_function_2)
        else:
            _app.callback(
                dash_outputs,
                dash_inputs,
                prevent_initial_call,
                running=dash_running_outputs,
                background=background,
            )(callback_function)

        return function

    return out_function


# set temporary files
cache = diskcache.Cache(PERSISTABLE_DASH_CACHE)
# cache_2 = diskcache.Cache(PERSISTABLE_DASH_CACHE+"_2")
background_callback_manager = DiskcacheManager(cache)
# long_callback_manager = DiskcacheLongCallbackManager(cache)
# background_callback_manager_2 = DiskcacheManager(cache_2)

_app = dash.Dash(
    __name__,
    # long_callback_manager=long_callback_manager,
    background_callback_manager=background_callback_manager,
)

min_granularity = 4
max_granularity = 8
default_log_granularity = 6

_app.layout = html.Div(
    children=[
        # contains the component counting function as a pandas dataframe
        dcc.Store(id=STORED_CCF),
        # contains the vineyard as a vineyard object
        dcc.Store(id=STORED_PV),
        html.Div(
            className="parameter-single",
            children=[
                html.Span(
                    className="name",
                    children="granularity comp count func",
                ),
                dcc.Slider(
                    min_granularity,
                    max_granularity,
                    step=None,
                    marks={i: str(2**i) for i in range(1, max_granularity + 1)},
                    value=default_log_granularity,
                    id=INPUT_LOG_GRANULARITY_CCF,
                    className=VALUE,
                ),
            ],
        ),
        html.Button(
            "(re)compute component counting function",
            id=COMPUTE_CCF_BUTTON,
            className=VALUE,
        ),
        html.Div(
            className="parameter-single",
            children=[
                html.Span(
                    className="name",
                    children="granularity vineyard",
                ),
                dcc.Slider(
                    min_granularity,
                    max_granularity,
                    step=None,
                    marks={i: str(2**i) for i in range(1, max_granularity + 1)},
                    value=default_log_granularity - 1,
                    id=INPUT_LOG_GRANULARITY_PV,
                    className=VALUE,
                ),
            ],
        ),
        html.Button(
            "compute prominence vineyard",
            id=COMPUTE_PV_BUTTON,
            className=VALUE,
        ),
    ],
)


def compute_ccf(d):
    print("CCF run")
    print(d)
    granularity = 2 ** d[INPUT_LOG_GRANULARITY_CCF + VALUE]
    import time

    time.sleep(2)

    return {STORED_CCF + DATA: json.dumps([])}

compute_ccf = my_callback(
    [
        [COMPUTE_CCF_BUTTON, N_CLICKS, IN],
        [INPUT_LOG_GRANULARITY_CCF, VALUE, ST],
    ],
    [
        [STORED_CCF, DATA],
    ],
    _app,
    prevent_initial_call=True,
    background=True,
    running=[[COMPUTE_CCF_BUTTON, DISABLED, True, False]],
)(compute_ccf)



def compute_pv(d):

    print("PV run")
    print(d)
    granularity = 2 ** d[INPUT_LOG_GRANULARITY_PV + VALUE]

    import time

    time.sleep(2)

    return {STORED_PV + DATA: json.dumps([])}
compute_pv = my_callback(
    [
        [COMPUTE_PV_BUTTON, N_CLICKS, IN],
        [INPUT_LOG_GRANULARITY_PV, VALUE, ST],
    ],
    [[STORED_PV, DATA]],
    _app,
    prevent_initial_call=True,
    background=True,
    running=[[COMPUTE_PV_BUTTON, DISABLED, True, False]],
    func_num=1)(compute_pv)


print(len(_funcs))

def PersistableInteractive(persistable, debug=False):
    _app.run_server(debug=debug)
